import re
import os
import argparse
import subprocess
import tempfile
import itertools
import shutil

import pysam

import handygenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.logutils as logutils
import handygenome.tools as tools
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.vcfeditor.varscan_editing as varscan_editing


PROGNAME = __name__.split(".")[-1]
JAVA_PATH = '/home/users/pjh/conda_bin/java'
VARSCAN2_JAR_PATH = '/home/users/tools/varscan2.4.2/VarScan.v2.4.2.jar'
SAMTOOLS_PATH = '/home/users/pjh/conda_bin/samtools'


def unit_job_1(fasta_path, regionfile_path, in_bam_path, out_pileup_path):
    logutils.print_timestamp('BEGINNING')

    logutils.print_timestamp('regionfile_path', regionfile_path)
    logutils.print_timestamp('in_bam_path', in_bam_path)
    logutils.print_timestamp('out_pileup_path', out_pileup_path)

    with open(regionfile_path) as infile:
        for line in infile:
            linesp = line.replace('\n', '').split('\t')
            chrom = linesp[0]
            start0 = int(linesp[1])
            end0 = int(linesp[2])
            start1 = start0 + 1
            end1 = end0

            args = [
                SAMTOOLS_PATH,
                'mpileup',
                '--no-BAQ',
                '-q', '0',
                '-Q', '0',
                '-f', f'{fasta_path}',
                '-r', f'{chrom}:{start1}-{end1}',
                '-o', f'{out_pileup_path}',
                f'{in_bam_path}',
            ]
            p = subprocess.run(args, capture_output=True, text=True)
            if p.returncode != 0:
                logutils.print_timestamp('Subprocess finished unsuccessfully. Stderr is:\n{p.stderr}')
                p.check_returncode()

    logutils.print_timestamp('FINISHED')


def unit_job_2(
    normal_pileup_path, 
    tumor_pileup_path, 
    vcf_prefix, 
    tumor_purity_opt,
    somatic_p_value_opt,
    min_coverage_normal_opt,
    min_coverage_tumor_opt,
):
    logutils.print_timestamp('BEGINNING')

    args = [
        JAVA_PATH, '-jar', VARSCAN2_JAR_PATH, 'somatic',
        normal_pileup_path,
        tumor_pileup_path,
        vcf_prefix,
        '--tumor-purity', str(tumor_purity_opt),
        '--somatic-p-value', str(somatic_p_value_opt),
        '--min-coverage-normal', str(min_coverage_normal_opt),
        '--min-coverage-tumor', str(min_coverage_tumor_opt),
        '--output-vcf', '1',
    ]
    p = subprocess.run(args, capture_output=True, text=True)
    if p.returncode != 0:
        logutils.print_timestamp('Subprocess finished unsuccessfully. Stderr is:\n{p.stderr}')
        p.check_returncode()

    logutils.print_timestamp('FINISHED')


def argument_parser(cmdargs):
    parser_dict = workflow.init_parser(
        description=(f"Runs gatk HaplotypeCaller with N parallel jobs.")
    )
    # required
    parser_dict['required'].add_argument(
        '--tbam',
        required=True,
        metavar='<tumor bam path>',
        dest='tbam_path',
        type=workflow.arghandler_infile,
        help=f'Tumor bam file path',
    )
    parser_dict['required'].add_argument(
        '--nbam',
        required=True,
        metavar='<normal bam path>',
        dest='nbam_path',
        type=workflow.arghandler_infile,
        help=f'Normal bam file path',
    )
    parser_dict['required'].add_argument(
        '--outfile-somatic',
        required=True,
        metavar='<somatic vcf path>',
        dest='somatic_outfile_path',
        type=workflow.get_arghandler_outfile(must_not_exist='ask'),
        help=f'Output VCF file path with "SOMATIC=TRUE" records.',
    )
    workflow.add_fasta_arg(parser_dict["required"], required=True)

    # optional
    parser_dict['optional'].add_argument(
        '--outfile-germline',
        required=False,
        metavar='<germline vcf path>',
        dest='germline_outfile_path',
        type=workflow.get_arghandler_outfile(must_not_exist='ask'),
        help=f'Output VCF file path with "SOMATIC=FALSE" records. If not set, germline VCF file is not written.',
    )

    workflow.add_outfmt_arg(parser_dict["optional"], required=False)
    workflow.add_incl_region_arg(parser_dict["optional"], required=False)
    workflow.add_excl_region_arg(parser_dict["optional"], required=False)
    workflow.add_logging_args(parser_dict)
    workflow.add_scheduler_args(parser_dict, default_parallel=1, default_sched="slurm")

    parser_dict["optional"].add_argument(
        '--java',
        required=False,
        dest='java_path',
        default=JAVA_PATH,
        help=f'Java executable path.',
    )
    parser_dict["optional"].add_argument(
        '--varscan-jar',
        required=False,
        dest='varscan_jar_path',
        default=VARSCAN2_JAR_PATH,
        help=f'VarScan2 jar file path.',
    )
    parser_dict["optional"].add_argument(
        '--tumor-purity',
        required=False,
        dest='tumor_purity_opt',
        default=1,
        type=float,
        help=f'"--tumor-purity" option given to "VarScan somatic" command. 1 is VarScan default.',
    )
    parser_dict["optional"].add_argument(
        '--somatic-p-value',
        required=False,
        dest='somatic_p_value_opt',
        default=0.05,
        type=float,
        help=f'"--somatic-p-value" option given to "VarScan somatic" command. 0.05 is VarScan default.',
    )
    parser_dict["optional"].add_argument(
        '--min-coverage-normal',
        required=False,
        dest='min_coverage_normal_opt',
        default=8,
        type=int,
        help=f'"--min-coverage-normal" option given to "VarScan somatic" command. 8 is VarScan default.',
    )
    parser_dict["optional"].add_argument(
        '--min-coverage-tumor',
        required=False,
        dest='min_coverage_tumor_opt',
        default=6,
        type=int,
        help=f'"--min-coverage-tumor" option given to "VarScan somatic" command. 8 is VarScan default.',
    )
    # flag
    workflow.add_rmtmp_arg(parser_dict)
    workflow.add_index_arg(parser_dict)
    # main
    args = parser_dict["main"].parse_args(cmdargs)
    return args


def make_tmpdir(infile_path):
    tmpdir_paths = workflow.get_tmpdir_paths(
        [
            "regions",
            "job1_scripts", "job1_logs", 
            "job2_scripts", "job2_logs", 
            "tumor_pileups", "normal_pileups",
            "split_results", 
        ],
        prefix=(os.path.basename(infile_path) + "_" + PROGNAME),
        where=os.path.dirname(infile_path),
    )

    return tmpdir_paths


def get_varscan2_version(java_path, varscan_jar_path):
    """Returns:
        VarScan2 version string, e.g. v2.4.2
    """
    p = subprocess.run(
        [java_path, '-jar', varscan_jar_path],
        capture_output=True,
        text=True,
    )
    return p.stderr.split('\n')[0].split()[1]


def write_job1_scripts(args, tmpdir_paths, num_split):
    script_path_list, log_path_list = toolsetup.get_script_log_paths(
        script_dir=tmpdir_paths['job1_scripts'], 
        log_dir=tmpdir_paths['job1_logs'], 
        num_split=(num_split * 2),
    )

    regionfile_path_list = list()
    for x in sorted(os.listdir(tmpdir_paths['regions'])):
        if 'padded' not in x:
            regionfile_path_list.append(os.path.join(tmpdir_paths['regions'], x))
    regionfile_path_list *= 2

    in_bam_path_list = (
        ([args.tbam_path] * num_split) 
        + ([args.nbam_path] * num_split)
    )

    out_pileup_path_list = list()
    for zidx in tools.zrange(num_split):
        out_pileup_path_list.append(os.path.join(tmpdir_paths['tumor_pileups'], f'{zidx}.pileup'))
    for zidx in tools.zrange(num_split):
        out_pileup_path_list.append(os.path.join(tmpdir_paths['normal_pileups'], f'{zidx}.pileup'))

    toolsetup.write_jobscripts(
        script_path_list=script_path_list,
        log_path_list=log_path_list,
        module_name=__name__,
        unit_job_func_name='unit_job_1',
        kwargs_single={
            'fasta_path': args.fasta_path,
        },
        kwargs_multi={
            'regionfile_path': regionfile_path_list,
            'in_bam_path': in_bam_path_list,
            'out_pileup_path': out_pileup_path_list,
        },
        jobname_prefix='VarScan2_wrapper_step1_makePileup',
        ncore_per_job=1,
    )

    return script_path_list


def write_job2_scripts(args, tmpdir_paths, num_split):
    script_path_list, log_path_list = toolsetup.get_script_log_paths(
        script_dir=tmpdir_paths['job2_scripts'], 
        log_dir=tmpdir_paths['job2_logs'], 
        num_split=num_split,
    )

    normal_pileup_path_list = list()
    tumor_pileup_path_list = list()
    vcf_prefix_list = list()
    for basename in sorted(os.listdir(tmpdir_paths['normal_pileups'])):
        normal_pileup_path_list.append(os.path.join(tmpdir_paths['normal_pileups'], basename))
    for basename in sorted(os.listdir(tmpdir_paths['tumor_pileups'])):
        tumor_pileup_path_list.append(os.path.join(tmpdir_paths['tumor_pileups'], basename))
    for npup_path, tpup_path in zip(normal_pileup_path_list, tumor_pileup_path_list):
        assert os.path.basename(npup_path) == os.path.basename(tpup_path)
        zidx = os.path.basename(npup_path).split('.')[0]
        vcf_prefix_list.append(os.path.join(tmpdir_paths['split_results'], zidx))

    toolsetup.write_jobscripts(
        script_path_list=script_path_list,
        log_path_list=log_path_list,
        module_name=__name__,
        unit_job_func_name='unit_job_2',
        kwargs_single={
            'tumor_purity_opt': args.tumor_purity_opt,
            'somatic_p_value_opt': args.somatic_p_value_opt,
            'min_coverage_normal_opt': args.min_coverage_normal_opt,
            'min_coverage_tumor_opt': args.min_coverage_tumor_opt,
        },
        kwargs_multi={
            'normal_pileup_path': normal_pileup_path_list,
            'tumor_pileup_path': tumor_pileup_path_list,
            'vcf_prefix': vcf_prefix_list,
        },
        jobname_prefix='VarScan2_wrapper_step2_runVarScan',
        ncore_per_job=1,
    )

    return script_path_list


def postprocess_and_merge(tmpdir_paths, args, fasta):
    infile_path_list = tools.listdir(tmpdir_paths['split_results'])
    varscan_editing.modify_varscan_vcf(
        infile_path_list, 
        somatic_outfile_path=args.somatic_outfile_path, 
        germline_outfile_path=args.germline_outfile_path,
        fasta=fasta,
        mode_pysam=args.mode_pysam, 
        pass_only=True,
        index=(not args.donot_index),
    )


def main(cmdargs):
    args = argument_parser(cmdargs)

    # setup parameters
    fasta = pysam.FastaFile(args.fasta_path)
    chromdict = refgenome.ChromDict.from_fasta(fasta)

    # make temporary directory tree
    tmpdir_paths = make_tmpdir(args.tbam_path)

    # setup logger
    logger = toolsetup.setup_logger(
        args=args, tmpdir_root=tmpdir_paths["root"], with_genlog=True,
    )
    logger.info("Beginning")

    # get VarScan version
    version = get_varscan2_version(args.java_path, args.varscan_jar_path)

    # make split intervals and write bed files
    toolsetup.handle_region_args(
        chromdict, 
        incl_bed_path=args.incl_bed_path,
        excl_bed_path=args.excl_bed_path,
        num_split=args.parallel,
        regionfiles_dir=tmpdir_paths["regions"],
    )

    # job1
    logger.info("Beginning job1 - making pileup for normal and tumor bam")
    script_path_list_job1 = write_job1_scripts(args=args, tmpdir_paths=tmpdir_paths, num_split=args.parallel)
    workflow.run_jobs(
        script_path_list_job1,
        sched=args.sched,
        intv_check=args.intv_check,
        intv_submit=args.intv_submit,
        logger=logger,
        log_dir=tmpdir_paths['job1_logs'],
        raise_on_failure=True,
    )
    logger.info("Finished job1")

    # job2
    logger.info("Beginning job2 - running VarScan")
    script_path_list_job2 = write_job2_scripts(args=args, tmpdir_paths=tmpdir_paths, num_split=args.parallel)
    workflow.run_jobs(
        script_path_list_job2,
        sched=args.sched,
        intv_check=args.intv_check,
        intv_submit=args.intv_submit,
        logger=logger,
        log_dir=tmpdir_paths['job2_logs'],
        raise_on_failure=True,
    )
    logger.info("Finished job2")
    
    # merge & clearance
    logger.info("Postprocess and merge")
    postprocess_and_merge(tmpdir_paths, args, fasta)

    if (not args.donot_rm_tmp):
        shutil.rmtree(tmpdir_paths['root'])

    logger.info("All successfully finished.")





import os
import tempfile
import subprocess

import Bio.SeqIO
import Bio.SeqRecord
import Bio.Seq
import pysam

import handygenome
import handygenome.network as network
import handygenome.refgenome as refgenome


BLAT_URL = 'https://genome.ucsc.edu/cgi-bin/hgBlat'
BWA = os.path.join(handygenome.DIRS['externals'], 'bwa')


def run_bwa(seqlist, refver, namelist=None):
    ref_path = refgenome.get_default_fasta_path(refver)
    if namelist is None:
        seqrec_list = [Bio.SeqRecord.SeqRecord(seq=Bio.Seq.Seq(seq), id='query')
                       for seq in seqlist]
    else:
        seqrec_list = list()
        for seq, seqname in zip(seqlist, namelist):
            seqrec = Bio.SeqRecord.SeqRecord(seq=Bio.Seq.Seq(seq), id=seqname)
            seqrec_list.append(seqrec)
    
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmpdir:
        query_path = os.path.join(tmpdir, 'input.fasta')
        sam_path = os.path.join(tmpdir, 'output.sam')
        Bio.SeqIO.write(seqrec_list, query_path, 'fasta')
        p = subprocess.run([
            BWA, 
            'mem', 
            '-Y', 
            '-M', 
            '-t', '1',
            '-o', sam_path,
            ref_path,
            query_path,
        ])
        readlist = list(pysam.AlignmentFile(sam_path).fetch())

    return readlist


def blat(query, refver):
    """Reference: https://genome.ucsc.edu/FAQ/FAQblat.html#blat14
    """
    refver = refgenome.RefverDict.converter[refver]

    result = network.http_get(
        url=BLAT_URL,
        params={
            'userSeq': query,
            'type': 'DNA',
            'db': refver,
            'output': 'json'
        }
    )
    
    if len(result['blat']) == 1:
        return dict(zip(result['fields'], result['blat'][0]))
    if len(result['blat'] == 0):
        return None
    elif len(result['blat'] > 1):
        raise Exception(f'Unexpected blat result:\n{result}')



#!/usr/bin/env python3
 
#Author: Erik Bergstrom

#Contact: ebergstr@eng.ucsd.edu

from __future__ import print_function
import os
import re
import sys
import argparse


def save_chrom_strings (genome, custom):
	'''
	Saves the fasta files for each chromosome as single strings. This
	is required to generate the reference files.

	Parameters:
		genome  -> reference genome

	Returns:
		None

	Ouput:
		Text files that contain single strings for each chromosome. Saved
		into the "/references/chromosomes/chrom_string/[genome]/" path
	'''

	# Instantiates all of the relevant paths/creates them if not present
	ref_dir, tail = os.path.split(os.path.dirname(os.path.abspath(__file__)))
	chrom_fasta_path = ref_dir + '/references/chromosomes/fasta/' + genome + "/"
	chrom_string_path = ref_dir + '/references/chromosomes/chrom_string/' + genome + '/'
	chrom_fasta_files = os.listdir(chrom_fasta_path)
	if not os.path.exists(chrom_string_path):
		os.makedirs(chrom_string_path)

	# Iterates through all of the fasta files for the given genome and saves
	# each into a single string.
	first = True
	i = 1
	if ".DS_Store" in chrom_fasta_files:
		chrom_fasta_files.remove('.DS_Store')
	chrom_total = len(chrom_fasta_files)
	for files in chrom_fasta_files:
		if not files.startswith('.'):
			file_name = files.split(".")
			try:
				chromosome = file_name[-2]
				if len(chromosome) > 2:
					chromosome = chromosome[3:]
			except:
				continue
			if custom:
				with open(chrom_fasta_path + files) as chrom, open(chrom_string_path + chromosome + ".txt", 'w') as out:
					next(chrom)
					for lines in chrom:
						line = lines.strip()
						print(line, file=out, end='')

					if first:
						print("The string file has been created for Chromosome: " + chromosome + " (" + str(i) + "/" + str(chrom_total) + ")")
						first = False
					else:
						print("                                                 " + chromosome + " (" + str(i) + "/" + str(chrom_total) + ")")
			else:
				if file_name[-4] == 'dna':
					with open(chrom_fasta_path + files) as chrom, open(chrom_string_path + chromosome + ".txt", 'w') as out:
						next(chrom)
						for lines in chrom:
							line = lines.strip()
							print(line, file=out, end='')

						if first:
							print("The string file has been created for Chromosome: " + chromosome + " (" + str(i) + "/" + str(chrom_total) + ")")
							first = False
						else:
							print("                                                 " + chromosome + " (" + str(i) + "/" + str(chrom_total) + ")")
		i += 1

def main():
	parser = argparse.ArgumentParser(description="Provide the necessary arguments to save the chromosomes as strings for later reference.")
	parser.add_argument("--genome", "-g",help="Provide a reference genome. (ex: GRCh37, GRCh38, mm10)")
	parser.add_argument("--custom", "-c",help="custom genome?. (True or False", default=False)
	args=parser.parse_args()
	genome = args.genome
	custom = args.custom

	save_chrom_strings(genome, custom)


if __name__ == '__main__':
	main()


import os
import itertools
import warnings

import pysam


def add_INFOmeta(vcfheader, ID, Type, Number, Description):
    vcfheader.add_meta(key='INFO',
                       items=[('ID', ID), ('Type', Type), ('Number', Number), 
                              ('Description', Description)])


def add_FORMATmeta(vcfheader, ID, Type, Number, Description):
    vcfheader.add_meta(key='FORMAT',
                       items=[('ID', ID), ('Type', Type), ('Number', Number), 
                              ('Description', Description)])


def add_generic(vcfheader, key, value):
    vcfheader.add_meta(key=key, value=value)


def add_structured(vcfheader, key, dic):
    vcfheader.add_meta(key=key, items=dic.items())


##################################


def addmeta_MATEID(vcfheader):
    if 'MATEID' not in vcfheader.info:
        vcfheader.add_meta(
            key='INFO',
            items=[('ID', 'MATEID'), ('Type', 'String'), ('Number', 1),
                   ('Description', 'ID of mate breakends')])


def addmeta_END(vcfheader):
    if 'END' not in vcfheader.info:
        vcfheader.add_meta(
            key='INFO',
            items=[('ID', 'END'), ('Type', 'Integer'), ('Number', 1),
                   ('Description', 'End position on CHROM')])


##################################


def merge_vcfheaders(vcfheader_list):
    """Any conflicting INFO or FORMAT keys (with regard to any of Type, 
    Number, or Description) are discarded.
    """
    
    # make a unified VariantHeader object which contains all INFO/FORMAT keys and samples
    vcfheader_list = list(vcfheader_list)
    merged_header = vcfheader_list[0].copy()
    for vcfheader in vcfheader_list[1:]:
        merged_header.merge(vcfheader)
        for sampleid in vcfheader.samples:
            if sampleid not in merged_header.samples:
                merged_header.add_sample(sampleid)

    # find conflicting keys
    info_metadatas = dict()
    format_metadatas = dict()
    for hdr in vcfheader_list:
        for key, val in hdr.info.items():
            info_metadatas.setdefault(key, set())
            info_metadatas[key].add((val.type, val.number, val.description))
        for key, val in hdr.formats.items():
            format_metadatas.setdefault(key, set())
            format_metadatas[key].add((val.type, val.number, val.description))

    conflicting_keys = {'info': set(), 'format': set()}
    for key, val in info_metadatas.items():
        if len(val) > 1:
            conflicting_keys['info'].add(key)
    for key, val in format_metadatas.items():
        if len(val) > 1:
            conflicting_keys['format'].add(key)

    # remove them
    for key in conflicting_keys['info']:
        merged_header.info.remove_header(key)
    for key in conflicting_keys['format']:
        merged_header.formats.remove_header(key)

    return merged_header, conflicting_keys


def merge_vcfheaders_heavy(vcfheader_list):
    """Any conflicting INFO or FORMAT keys 
    (with regard to any of Type, Number, or Description) are discarded."""
    
    vcfheader_list = list(vcfheader_list)
    result = vcfheader_list[0].copy()
    for vcfheader in vcfheader_list[1:]:
        result.merge(vcfheader)
        for sampleid in vcfheader.samples:
            if sampleid not in result.samples:
                result.add_sample(sampleid)

    # get conflicting keys
    conflicting_keys_sum = {'info': set(), 'format': set()}
    for vcfheader1, vcfheader2 in itertools.combinations(vcfheader_list, 2):
        compatible, conflicting_keys = check_header_compatibility(
            vcfheader1, vcfheader2)
        conflicting_keys_sum['info'].update(conflicting_keys['info'])
        conflicting_keys_sum['format'].update(conflicting_keys['format'])

    # remove them
    for key in conflicting_keys_sum['info']:
        result.info.remove_header(key)
    for key in conflicting_keys_sum['format']:
        result.formats.remove_header(key)

    return result


def check_header_compatibility(vcfheader1, vcfheader2):
    conflicting_keys = {'info': list(), 'format': list()}

    info_keys_1 = set(vcfheader1.info.keys())
    info_keys_2 = set(vcfheader2.info.keys())
    info_keys_isec = info_keys_1.intersection(info_keys_2)
    for key in info_keys_isec:
        meta1 = vcfheader1.info[key]
        meta2 = vcfheader2.info[key]
        for meta_key in ('name', 'type', 'number', 'description'):
            if getattr(meta1, meta_key) != getattr(meta2, meta_key):
                conflicting_keys['info'].append(key)
                break

    format_keys_1 = set(vcfheader1.formats.keys())
    format_keys_2 = set(vcfheader2.formats.keys())
    format_keys_isec = format_keys_1.intersection(format_keys_2)
    for key in format_keys_isec:
        meta1 = vcfheader1.formats[key]
        meta2 = vcfheader2.formats[key]
        for meta_key in ('name', 'type', 'number', 'description'):
            if getattr(meta1, meta_key) != getattr(meta2, meta_key):
                conflicting_keys['format'].append(key)
                break

    compatible = (len(conflicting_keys['info']) == 0 and 
                  len(conflicting_keys['format']) == 0)

    return compatible, conflicting_keys


#def check_header_compatibility_list(pysamhdr_list):
#    compatible_list = list()
#    for pysamhdr1, pysamhdr2 in itertools.combinations(pysamhdr_list, 2):
#        compatible, differing_keys = check_header_compatibility_two(pysamhdr1, pysamhdr2)
#        compatible_list.append(compatible)
#    return all(compatible_list)
#
#
#def check_header_compatibility_two(pysamhdr1, pysamhdr2):
#    differing_keys = dict()
#    differing_keys['info_type'] = list()
#    differing_keys['info_number'] = list()
#    differing_keys['format_type'] = list()
#    differing_keys['format_number'] = list()
#
#    info_keys_1 = set(pysamhdr1.info.keys())
#    info_keys_2 = set(pysamhdr2.info.keys())
#    info_keys_isec = info_keys_1.intersection(info_keys_2)
#    for key in info_keys_isec:
#        meta1 = pysamhdr1.info[key]
#        meta2 = pysamhdr2.info[key]
#        if meta1.type != meta2.type:
#            differing_keys['info_type'].append(key)
#        if meta1.number != meta2.number:
#            differing_keys['info_number'].append(key)
#
#    format_keys_1 = set(pysamhdr1.formats.keys())
#    format_keys_2 = set(pysamhdr2.formats.keys())
#    format_keys_isec = format_keys_1.intersection(format_keys_2)
#    for key in format_keys_isec:
#        meta1 = pysamhdr1.formats[key]
#        meta2 = pysamhdr2.formats[key]
#        if meta1.type != meta2.type:
#            differing_keys['format_type'].append(key)
#        if meta1.number != meta2.number:
#            differing_keys['format_number'].append(key)
#
#    compatible = all(len(x) == 0 for x in differing_keys.values())
#
#    return compatible, differing_keys



def close_vcfplus_list(vcfplus_list):
    for vcfp in vcfplus_list:
        vcfp.vcf.close()




import collections
import operator

import Bio.Align


DEFAULT_FLANKLEN = 1

ALIGNER_COMPARE = Bio.Align.PairwiseAligner(
    mode='global',
    match_score=0,
    mismatch_score=-1,
    query_gap_score=-1,
    target_gap_score=-1,
)


"""Meaning of alleleclass:
    None: not informative. When a read does not span flanking areas.
        All other alleleclasses imply a read spans flanking areas.
    -1: other than REF and ALTS
    0: REF
    1: 1st ALT
    2: 2nd ALT
    ...
"""


# merging results of two readplus objects

def merge_monoalt_supports(supports_rp1, supports_rp2):
    """Args:
        results of "get_normalized_monoalt_supports_readplus" function
    """
    merged_monoalt_supports = dict()
    for alt_index in supports_rp1.keys():
        merged_monoalt_supports[alt_index] = merge_alleleclasses(
            supports_rp1[alt_index], supports_rp2[alt_index],
        )

    return merged_monoalt_supports


def merge_alleleclasses(alleleclass_rp1, alleleclass_rp2):
    if (alleleclass_rp1 is not None) and (alleleclass_rp2 is not None):
        if alleleclass_rp1 == alleleclass_rp2:
            return alleleclass_rp1
        else:
            return None
    elif (alleleclass_rp1 is None) and (alleleclass_rp2 is None):
        return None
    else:
        if alleleclass_rp1 is not None:
            return alleleclass_rp1
        elif alleleclass_rp2 is not None:
            return alleleclass_rp2


#############################################

# asis

def get_alleleclass_asis_readplus_new(
    vcfspec, rp, 
    coverage_cutoff=1, 
    partial_coverage_del_len_cutoff=5,
    compare_mode='identity', 
    similarity_params={'factor': 0.1},
):
    """Each ReadPlus can support no more than one allele"""

    #assert compare_mode in ('identity', 'similarity')

    # cigar N full span check
    if rp.check_cigarN_includes_range(vcfspec.range0):
        return None

    # treat as None if not spanning relevant repeat region
    read_range0 = rp.get_range0()
    spans_repeat_list = list()
    for repeat_spec in vcfspec.repeat_specs:
        if repeat_spec is None:
            spans_repeat = True
        else:
            spans_repeat = (
                (repeat_spec.start0 in read_range0)
                and (repeat_spec.end0 in read_range0)  
            )  # The position 1 base right to the repeat end must be spanned by the read
        spans_repeat_list.append(spans_repeat)

    if not all(spans_repeat_list):
        return None

    # vcfspec coverage
    spans_left, spans_right, relevant_read_seq = get_vcfspec_coverage_info(rp, vcfspec)

    # treat as None by vcfspec coverage
    if (not spans_left) and (not spans_right):  # no span
        return None
    if (not spans_left) or (not spans_right):  # partial span
        vcfspec_coverage, modified_vcfspec_range0 = get_vcfspec_coverage_fraction(vcfspec, rp)
        if len(modified_vcfspec_range0) >= partial_coverage_del_len_cutoff:
            # len(modified_vcfspec_range0) equals to the deletion length
            if not (vcfspec_coverage >= coverage_cutoff):
                return None
        else:
            return None
        
    # get seqs to use as target (read sequence is query)
    target_seqs = get_target_seqs(vcfspec, spans_left, spans_right, relevant_read_seq)
    # compare target and query seqs
    alleleclass = compare_targets_and_query(
        target_seqs, relevant_read_seq, mode=compare_mode, similarity_params=similarity_params,
    )

    return alleleclass


def get_alleleclass_asis_readplus(vcfspec, rp, flanklen=DEFAULT_FLANKLEN):
    """Each ReadPlus can support no more than one allele"""

    assert flanklen >= 1, f'"flanklen" argument must be at least 1.'

    # cigar N full span check
    if rp.check_cigarN_includes_range(vcfspec.range0):
        return None

    spans, matches = rp.check_spans_and_matches_vcfspec_flanks(vcfspec, flanklen=flanklen)
    if spans:
        if matches:
            allele_seq = rp.get_seq_from_range0(
                vcfspec.REF_range0,
                flanking_queryonly_default_mode=False,
                include_leading_queryonly=True, 
                include_trailing_queryonly=True,
            )

            if allele_seq in vcfspec.alleles:
                alleleclass = vcfspec.alleles.index(allele_seq)
            else:
                alleleclass = -1
        else:
            alleleclass = -1
    else:
        if 'S' in rp.read.cigarstring:
            if rp.check_softclip_spans_vcfspec_flanks(vcfspec, flanklen=flanklen):
                alleleclass = -1
            else:
                alleleclass = None
        else:
            alleleclass = None

    return alleleclass


################################################################

# helpers

def alleleclass_sortkey(x):
    if x is None:
        return -2
    else:
        return x


def get_vcfspec_coverage_info(rp, vcfspec):
    if rp.check_all_queryonly():
        spans_left = False
        spans_right = False
        relevant_read_seq = None
    else:
        read_range0 = rp.range0
        spans_left = (min(vcfspec.readspan_range0) in read_range0)
        spans_right = (max(vcfspec.readspan_range0) in read_range0)
        relevant_read_seq = rp.get_seq_from_pairs_indexes(
            rp.get_pairs_indexes(
                vcfspec.readspan_range0, flanking_queryonly_default_mode=True,
            )
        )

    return spans_left, spans_right, relevant_read_seq


def get_vcfspec_coverage_fraction(vcfspec, rp):
    mutation_types = set(
        vcfspec.get_mutation_type(alt_index=alt_index)
        for alt_index in vcfspec.iter_alt_indexes()
    )
    if mutation_types == {'del'}:
        modified_vcfspec_range0 = range(
            vcfspec.readspan_range0.start + 1,
            vcfspec.readspan_range0.stop - 1,
        )
    else:
        modified_vcfspec_range0 = vcfspec.readspan_range0

    read_span_start = max(rp.range0.start, modified_vcfspec_range0.start)
    read_span_stop = min(rp.range0.stop, modified_vcfspec_range0.stop)
    numerator = read_span_stop - read_span_start
    denominator = len(modified_vcfspec_range0)
    vcfspec_coverage = numerator / denominator

    return vcfspec_coverage, modified_vcfspec_range0


def get_target_seqs(vcfspec, spans_left, spans_right, relevant_read_seq):
    target_seqs = vcfspec.compared_seq_dict.copy()
    if spans_left and (not spans_right):
        target_seqs = {key: val[:len(relevant_read_seq)] for key, val in target_seqs.items()}
    elif (not spans_left) and spans_right:
        target_seqs = {key: val[-len(relevant_read_seq):] for key, val in target_seqs.items()}

    return target_seqs


def compare_targets_and_query(target_seqs, relevant_read_seq, mode, similarity_params):
    """Args:
        target_seqs: Result of Vcfspec.compared_seq_dict method. 
            keys are allele_index.
    Result:
        alleleclass
    """
    assert mode in ('identity', 'similarity')

    if mode == 'identity':
        hit_allele_indexes = list()
        for allele_index, target in target_seqs.items():
            if target == relevant_read_seq:
                hit_allele_indexes.append(allele_index)

        if len(hit_allele_indexes) == 1:
            return hit_allele_indexes[0]
        else:
            return None

    elif mode == 'similarity':
        alignment_info = list()
        for allele_index, target in target_seqs.items():
            alns = ALIGNER_COMPARE.align(target, relevant_read_seq)
            if similarity_checker(alns, **similarity_params):
                alignment_info.append((allele_index, alns.score))
        if len(alignment_info) == 0:
            return None
        else:
            alignment_info.sort(key=operator.itemgetter(1))
            return alignment_info[-1][0]


def similarity_checker(alns, factor=0.1):
    """Returns:
        True means passed
    """
    query_len = len(alns.sequences[1])
    return -alns.score < (query_len * factor)
            

################################################################

# monoalt support

class VariantSupportMonoALT(
    collections.namedtuple(
        'VariantSupport', 
        ('supports', 'relevant')
    )
):
    pass


def check_monoalt_support(vcfspec, rp, flanklen=DEFAULT_FLANKLEN):
    alleleclass = get_alleleclass_asis_readplus(vcfspec, rp, flanklen=flanklen)
    if alleleclass is None:
        return None
    else:
        return (alleleclass == 1)


def get_normalized_monoalt_supports_readplus(vcfspec, rp, flanklen=DEFAULT_FLANKLEN):
    monoalt_supports = dict()
    monoalt_list = [x.normalize() for x in vcfspec.merge_components]
    #for alt_index, vcfspec_monoalt in enumerate(monoalt_list):
    for alt_index, vcfspec_monoalt in enumerate(monoalt_list):
        monoalt_supports[alt_index] = check_monoalt_support(vcfspec_monoalt, rp, flanklen=flanklen)
        #monoalt_supports.append(check_monoalt_support(vcfspec_monoalt, rp, flanklen=flanklen))

    return monoalt_supports





import functools

import pyranges as pr

import handygenome.tools as tools
import handygenome.deco as deco


class Interval:
    """
    Attributes:
        chrom
        start0
        end0
        start1
        end1
        range0
        length
    """

    def __init__(
        self, chrom, *, start1=None, end1=None, start0=None, end0=None, is_reverse=False,
    ):
        """Args:
            'chrom' is mandatory.
            ('start1' and 'end1') or ('start0' and 'end0') must 
                be given (for coordinate setting).
            'start0' and 'end0' are 0-based half-open system.
            'start1' and 'end1' are 1-based closed system.
        """

        self.chrom = chrom
        self.is_reverse = is_reverse

        # start0, end0, start1, end1
        if (start1 is not None) and (end1 is not None):
            self.start0 = start1 - 1
            self.end0 = end1
        else:
            self.start0 = start0
            self.end0 = end0

        self.start1 = self.start0 + 1
        self.end1 = self.end0

        # range
        self.range0 = range(self.start0, self.end0)

        # length
        self.length = self.end0 - self.start0

    def __repr__(self):
        return (f'<Interval> {self.chrom}:{self.start1:,}-{self.end1:,} '
                f'(1-based)')

    def to_gr(self, **kwargs):
        data = {
            'Chromosome': [self.chrom],
            'Start': [self.start0],
            'End': [self.end0],
        }
        for key, val in kwargs.items():
            data[key] = [val]
        return pr.from_dict(data)

    @classmethod
    def from_gr(cls, gr):
        assert len(gr) == 1
        return cls(chrom=gr.Chromosome[0], start0=gr.Start[0], end0=gr.End[0])

    def includes(self, other):
        return (self.chrom == other.chrom and
                self.start0 <= other.start0 and
                self.end0 >= other.end0)


class IntervalList(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lengths_cumsum = None

    # constructors #
    @classmethod
    def from_gr(cls, gr):
        result = cls()
        for chrom, start0, end0 in zip(gr.Chromosome, gr.Start, gr.End):
            intv = Interval(chrom=chrom, start0=start0, end0=end0)
            result.append(intv)
        return result

    @classmethod
    def from_bed(cls, bedfile):
        gr = pr.read_bed(bedfile)
        return cls.from_gr(gr)

    @classmethod
    def from_chromdict(cls, chromdict):
        result = cls()
        for contig, length in chromdict.items():
            result.append(Interval(chrom=contig, start0=0, end0=length))

        return result

    @classmethod
    def from_vcfspec(cls, vcfspec):
        return cls.from_gr(cls, vcfspec.to_gr())

    @classmethod
    def from_margin(cls, chromdict, chrom_left, start0_left, chrom_right, end0_right):
        #chromdict = DEFAULT_CHROMDICTS[refver]
        cumpos0_left = chromdict.get_cumpos0(chrom_left, start0_left)
        cumpos0_right = chromdict.get_cumpos0(chrom_right, end0_right)
        if cumpos0_left >= cumpos0_right:
            raise Exception(
                f'"left position" comes later than "right_position"; '
                f'chrom_left={chrom_left}, start0_left={start0_left}, '
                f'chrom_right={chrom_right}, end0_right={end0_right}'
            )

        result = cls()
        if chrom_left == chrom_right:
            result.append(Interval(chrom=chrom_left, start0=start0_left, end0=end0_right))
        else:
            chrom_left_idx = chromdict.contigs.index(chrom_left)
            chrom_right_idx = chromdict.contigs.index(chrom_right)
            result.append(
                Interval(chrom=chrom_left, start0=start0_left, end0=chromdict[chrom_left])
            )
            for chrom in chromdict.contigs[(chrom_left_idx + 1):chrom_right_idx]:
                result.append(
                    Interval(chrom=chrom, start0=0, end0=chromdict[chrom])
                )
            result.append(
                Interval(chrom=chrom_right, start0=0, end0=end0_right)
            )

        return result

    @classmethod
    def get_depth_bins(cls, chromdict, width=100_000):
        result = cls()
        for contig, length in zip(chromdict.contigs, chromdict.lengths):
            intvlist_contig = cls()
            intvlist_contig.append(Interval(chrom=contig, start0=0, end0=length))
            for item in intvlist_contig.split(width=width):
                result.extend(item)
        return result
    ################

    def write_bed(self, outfile_path):
        with tools.openfile(outfile_path, 'w') as outfile:
            for intv in self:
                outfile.write(f'{intv.chrom}\t{intv.start0}\t{intv.end0}\n')

    def to_gr(self):
        chroms = list()
        starts = list()
        ends = list()
        for intv in self:
            chroms.append(intv.chrom)
            starts.append(intv.start0)
            ends.append(intv.end0)
        return pr.from_dict({'Chromosome': chroms, 'Start': starts, 'End': ends})

    # properties
    @functools.cached_property
    def lengths_cumsum(self):
        return list(itertools.accumulate(intv.length for intv in self))
    @property
    def length(self):
        return sum(intv.length for intv in self)

    # inclusion check
    def includes_vr(self, vr):
        vr_intv = Interval(chrom=vr.contig, start1=vr.pos, end1=vr.pos)
        return any(intv.includes(vr_intv) for intv in self)

    # calculations
    def sort_intervals(self, chromdict):
        def sortkey(intv):
            return tools.coord_sortkey(intv.chrom, intv.start1, chromdict)
        self.sort(key=sortkey)

    def isec(self, other):
        self_gr = self.to_gr()
        other_gr = other.to_gr()
        isec_gr = self_gr.set_intersect(other_gr)

        return self.__class__.from_gr(isec_gr)

    def subtract(self, other):
        self_gr = self.to_gr()
        other_gr = other.to_gr()
        subtract_gr = self_gr.subtract(other_gr)

        return self.__class__.from_gr(subtract_gr)

    def union(self, other):
        self_gr = self.to_gr()
        other_gr = other.to_gr()
        union_gr = self_gr.set_union(other_gr)

        return self.__class__.from_gr(union_gr)

    def merge(self):
        return self.__class__.from_gr(self.to_gr().merge())

    @deco.get_deco_num_set(('b', 'l', 'r'), 1)
    def slop(self, chromdict, b=None, l=None, r=None):
        def start_handler(start0, width):
            new_start0 = max(0, start0 - width)
            return new_start0

        def end_handler(end0, width, chrom, chromdict):
            new_end0 = min(chromdict[chrom], end0 + width)
            return new_end0

        result = self.__class__()
        for intv in self:
            if b is not None:
                new_start0 = start_handler(intv.start0, b)
                new_end0 = end_handler(intv.end0, b, intv.chrom, chromdict)
            elif l is not None:
                new_start0 = start_handler(intv.start0, l)
                new_end0 = intv.end0
            elif r is not None:
                new_start0 = new_start0
                new_end0 = end_handler(intv.end0, b, intv.chrom, chromdict)

            new_intv = Interval(chrom=intv.chrom, start0=new_start0, end0=new_end0)
            result.append(new_intv)

        return result

    @deco.get_deco_num_set(('num', 'width'), 1)
    def split(self, *, num=None, width=None):
        #self.sort_intervals(chromdict)

        # get result_lengths_cumsum
        total_length = self.lengths_cumsum[-1]
        if num is not None:
            result_lengths_cumsum = list(
                itertools.accumulate(get_interval_lengths_num(total_length, num))
            )
        elif width is not None:
            result_lengths_cumsum = list(itertools.accumulate(
                get_interval_lengths_width(total_length, width)))

        # prepare result values
        result = list()
        for idx in range(len(result_lengths_cumsum)):
            # get start0 and end0 in the single merged coordinate system
            merged_start0 = (0 
                             if idx == 0 else
                             result_lengths_cumsum[idx - 1])
            merged_end0 = result_lengths_cumsum[idx]
            # modify merged coordinates into interval-specific ones
            (chrom_start, 
             pos0_start, 
             self_idx_start) = self._modify_coord(merged_start0, is_end=False)
            (chrom_end, 
             pos0_end, 
             self_idx_end) = self._modify_coord(merged_end0, is_end=True)
            # create an IntervalList corresponding to a split unit
            intvlist = IntervalList()
            if self_idx_start == self_idx_end:
                intv = Interval(chrom_start, start0=pos0_start, end0=pos0_end)
                intvlist.append(intv)
            else:
                intv_first = Interval(chrom_start, 
                                      start0=pos0_start, 
                                      end0=self[self_idx_start].end0)
                intvlist.append(intv_first)

                for self_idx in range(self_idx_start + 1, self_idx_end):
                    intvlist.append(self[self_idx])

                intv_last = Interval(chrom_end, 
                                     start0=self[self_idx_end].start0, 
                                     end0=pos0_end)
                intvlist.append(intv_last)

            result.append(intvlist)

        return result

    def _modify_coord(self, merged_pos0, is_end=False):
        def get_interval_values(merged_pos0, idx, self):
            chrom = self[idx].chrom

            if idx == 0:
                shift_within_interval = merged_pos0
            else:
                shift_within_interval = (merged_pos0 
                                         - self.lengths_cumsum[idx - 1])
            new_pos0 = self[idx].start0 + shift_within_interval

            return chrom, new_pos0

        def handle_current_interval(merged_pos0, idx, length_previous, 
                                    length_current, is_end):
            if merged_pos0 > length_previous and merged_pos0 <= length_current:
                if merged_pos0 == length_current:
                    if is_end:
                        chrom, new_pos0 = get_interval_values(
                            merged_pos0, idx, self)
                    else:
                        chrom, new_pos0 = get_interval_values(
                            merged_pos0, idx + 1, self)
                elif merged_pos0 < length_current:
                    chrom, new_pos0 = get_interval_values(merged_pos0, idx, 
                                                          self)

                to_break = True
            else:
                chrom = None
                new_pos0 = None
                to_break = False

            return chrom, new_pos0, to_break

        # sanity check
        assert merged_pos0 >= 0, f'"merged_pos0" must be non-negative.'
        assert not ((not is_end) and merged_pos0 >= self.lengths_cumsum[-1]), (
            f'If ""is_end" is False, "merged_pos0" must be less than '
            f'the total IntervalList length.')
        assert not (is_end and merged_pos0 > self.lengths_cumsum[-1]), (
            f'If ""is_end" is True, "merged_pos0" must be less than or '
            f'equal to the total IntervalList length.')

        if merged_pos0 == 0:
            chrom = self[0].chrom
            new_pos0 = self[0].start0
            self_idx = 0
        else:
            for idx in range(len(self.lengths_cumsum)):
                length_current = self.lengths_cumsum[idx]
                length_previous = (0 
                                   if idx == 0 else
                                   self.lengths_cumsum[idx - 1])
                (chrom, new_pos0, to_break) = handle_current_interval(
                    merged_pos0, idx, length_previous, length_current, is_end)
                self_idx = idx

                if to_break:
                    break

        return chrom, new_pos0, self_idx


# interval-related functions

def get_interval_lengths_num(total_length, num):
    interval_num = min(num, total_length)
    q, r = divmod(total_length, interval_num)
    interval_width_1 = q + 1
    interval_num_1 = r
    interval_width_2 = q
    interval_num_2 = interval_num - r

    result = list(
        itertools.chain(
            itertools.repeat(interval_width_1, interval_num_1),
            itertools.repeat(interval_width_2, interval_num_2),
        )
    )

    return result


def get_interval_lengths_width(total_length, width):
    interval_width_raw = min(width, total_length)
    q, r = divmod(total_length, interval_width_raw)
    if r == 0:
        interval_width = interval_width_raw
        interval_num = q
        result = list(itertools.repeat(interval_width, interval_num))
    else:
        q2, r2 = divmod(r, q)
        interval_width_1 = interval_width_raw + q2 + 1
        interval_num_1 = r2
        interval_width_2 = interval_width_raw + q2
        interval_num_2 = q - r2
        result = list(itertools.chain(
            itertools.repeat(interval_width_1, interval_num_1),
            itertools.repeat(interval_width_2, interval_num_2)))

    return result


# range operation

def range_included(rng1, rng2):
    """Returns:
        True if rng1 is included in rng2
    """
    return (rng1.start >= rng2.start) and (rng1.stop <= rng2.stop)


def range_intersection(rng1, rng2):
    start = max(rng1.start, rng2.start)
    stop = min(rng1.stop, rng2.stop)
    if stop > start:
        return range(start, stop)
    else:
        return None


def range_subtraction(rng1, rng2):
    isec = range_intersection(rng1, rng2)
    if isec is None:
        return rng1
    else:
        if isec.start == rng1.start:
            if isec.stop == rng1.stop:
                return None
            else:
                return range(isec.stop, rng1.stop)
        else:
            if isec.stop == rng1.stop:
                return range(rng1.start, isec.start)
            else:
                return (range(rng1.start, isec.start), range(isec.stop, rng1.stop))


def reverse_range(rng):
    if rng.step == 1:
        start = rng.stop - 1
        stop = rng.start - 1
        return range(start, stop, -1)
    elif rng.step == -1:
        start = rng.stop + 1
        stop = rng.start + 1
        return range(start, stop, 1)
    else:
        raise Exception(f'Input range obejct must have "step" value of either 1 or -1.')


def check_overlaps_forward_nonzero(rng1, rng2):
    return (rng1.start < rng2.stop) and (rng1.stop > rng2.start)


def check_overlaps(rng1, rng2):
    """
    Meaning of range
        - Assumes range.step is either 1 or -1
        - "current position" travels, beginning from "range.start" toward "range.stop".
        - "cursor" position
            - When range.step == 1: immediately left to "current position"
            - When range.step == -1: immediately right to "current position"
        - Area swept by "cursor" is the area represented by range object.
        - When len(range) == 0:
            - When range.step == 1: 0-length interval between (range.start - 1) and (range.start)
            - When range.step == -1: 0-length interval between (range.start) and (range.start + 1)

    Meaning of overlap with 0-length range
        - consider as overlap when the interspace is contained between ALIGNED read bases (not insertion)

    Args:
        rng1, rng2: python range object.
    """
    # convert into positive-step range
    if rng1.step == -1:
        rng1 = reverse_range(rng1)
    if rng2.step == -1:
        rng2 = reverse_range(rng2)

    # main
    if len(rng1) == 0 and len(rng2) == 0:
        return rng1.start == rng2.start
    else:
        return check_overlaps_forward_nonzero(rng1, rng2)


def check_spans(query_range0, target_range0):
    """Checks if query spans target"""

    if len(query_range0) == 0:
        return False
    else:
        if len(target_range0) == 0:
            return check_overlaps(query_range0, target_range0)
        else:
            return (
                min(query_range0) <= min(target_range0) and
                max(query_range0) >= max(target_range0)
            )





import re
import itertools
import collections
import functools

import numpy as np
import pandas as pd
import pyranges as pr

import importlib

top_package_name = __name__.split(".")[0]
common = importlib.import_module(".".join([top_package_name, "common"]))
readhandler = importlib.import_module(
    ".".join([top_package_name, "read", "readhandler"])
)
alignhandler = importlib.import_module(".".join([top_package_name, "align", "alignhandler"]))
librealign = importlib.import_module(".".join([top_package_name, "align", "realign"]))
fetchcache = importlib.import_module(".".join([top_package_name, "read", "fetchcache"]))


DEL_VALUE = "*"
EMPTY_VALUE = ""
DEFAULT_EXTEND_FETCHEDREADS_BY = 300
DEFAULT_EXTEND_PILEUP_BY = 30
#MAX_READ_LENGTH = 151
#
#
#def fetch_arghandler(query_start0, query_end0, db_start0, db_end0):
#    if query_start0 is None:
#        result_start0 = db_start0
#    else:
#        result_start0 = max(query_start0, db_start0)
#
#    if query_end0 is None:
#        result_end0 = db_end0
#    else:
#        result_end0 = min(query_end0, db_end0)
#
#    return result_start0, result_end0
#
#
#class FetchedReadsCluster:
#    def __init__(self, bam, chrom, readfilter=None):
#        self.bam = bam
#        self.chrom = chrom
#
#        if readfilter is None:
#            self.readfilter = readhandler.readfilter_pileup
#        else:
#            self.readfilter = readfilter
#
#        self.fetch_start0 = None
#        self.fetch_end0 = None
#        self.dict = collections.OrderedDict()
#
#    def __repr__(self):
#        return f'<FetchedReadsCluster object (chrom={self.chrom}, fetch_start0={self.fetch_start0:,}, fetch_end0={self.fetch_end0:,})>'
#
#    @classmethod
#    def from_fetch(cls, bam, chrom, start0, end0, readfilter=None):
#        result = cls(bam, chrom, readfilter)
#        result.fetch_start0 = start0
#        result.fetch_end0 = end0
#
#        for read in readhandler.get_fetch(
#            result.bam, result.chrom, start0, end0, readfilter=result.readfilter
#        ):
#            read_uid = readhandler.get_uid(read)
#            result.dict[read_uid] = read
#
#        return result
#
#    ###########
#    # yielder #
#    ###########
#    @staticmethod
#    def yielder_with_uid(uid, read):
#        return uid, read
#
#    @staticmethod
#    def yielder_without_uid(uid, read):
#        return read
#
#    def get_yielder(self, with_uid):
#        if with_uid:
#            return self.yielder_with_uid
#        else:
#            return self.yielder_without_uid
#    ###########
#
#    @property
#    def fetch_range0(self):
#        return range(self.fetch_start0, self.fetch_end0)
#
#    @property
#    def readspan_start0(self):
#        if len(self.dict) == 0:
#            return None
#        else:
#            return min(read.reference_start for read in self.dict.values())
#
#    @property
#    def readspan_end0(self):
#        if len(self.dict) == 0:
#            return None
#        else:
#            return max(read.reference_end for read in self.dict.values())
#
##    def _check_uid_uniqueness(self):
##        if len(set(self.uids)) != len(self.uids):
##            reads_string = "\n".join(x.to_string() for x in self.list)
##            uids_string = "\n".join(x for x in self.uids)
##            raise Exception(
##                f"Duplicate read uids.\n"
##                f"Read list:\n{reads_string}.\n"
##                f"uid list:\n{uids_string}."
##            )
#
#    def get_id(self):
#        return (self.chrom, self.fetch_start0, self.fetch_end0)
#
#    def get_read(self, read_uid):
#        return self.dict[read_uid]
#
#    def copy(self):
#        result = self.__class__(self.bam, self.chrom)
#        result.fetch_start0 = self.fetch_start0
#        result.fetch_end0 = self.fetch_end0
#        result.dict.update(self.dict)
#        return result
#
#    # merge, extend #
#    def merge(self, other):
#        if (other.fetch_start0 > self.fetch_end0) or (self.fetch_start0 > other.fetch_end0):
#            raise Exception(f'Fetch ranges of two objects do not overlap.')
#
#        if self.fetch_start0 <= other.fetch_start0 and self.fetch_end0 >= other.fetch_end0:
#            return self.copy()
#        elif other.fetch_start0 <= self.fetch_start0 and other.fetch_end0 >= self.fetch_end0:
#            return other.copy()
#        else:
#            # now self.fetch_start0 != other.fetch_start0
#            result = self.__class__(self.bam, self.chrom)
#            if self.fetch_start0 < other.fetch_start0:
#                left = self
#                right = other
#                result.fetch_start0 = self.fetch_start0
#                result.fetch_end0 = other.fetch_end0
#            else:
#                left = other
#                right = self
#                result.fetch_start0 = other.fetch_start0
#                result.fetch_end0 = self.fetch_end0
#
#            left_only_width = right.fetch_start0 - left.fetch_start0
#            right_only_width = right.fetch_end0 - left.fetch_end0
#            if left_only_width >= right_only_width:
#                result.dict.update(left.dict)
#
#                buffer = list(
#                    itertools.takewhile(
#                        (lambda x: x[0] not in left.dict),
#                        reversed(right.dict.items())
#                    )
#                )
#                result.dict.update(reversed(buffer))
#            else:
#                result.dict.update(
#                    itertools.takewhile(
#                        (lambda x: x[0] not in right.dict),
#                        left.dict.items()
#                    )
#                )
#                result.dict.update(right.dict)
#
#            return result
#
#    def extend_rightward(self, width):
#        for read, uid in itertools.dropwhile(
#            (lambda x: x[1] in self.dict),
#            readhandler.get_fetch(self.bam, self.chrom, self.fetch_end0, self.fetch_end0 + width, readfilter=self.readfilter, with_uid=True),
#        ):
#            self.dict[uid] = read
#
#        self.fetch_end0 += width
#
#    def extend_leftward(self, width):
#        buffer = list(
#            itertools.takewhile(
#                (lambda x: x[1] not in self.dict),
#                readhandler.get_fetch(self.bam, self.chrom, self.fetch_start0 - width, self.fetch_start0, readfilter=self.readfilter, with_uid=True),
#            )
#        )
#        for read, uid in reversed(buffer):
#            self.dict[uid] = read
#            self.dict.move_to_end(uid, last=False)
#
#        self.fetch_start0 -= width
#            
#    # fetch #
#    def fetch(self, start0=None, end0=None, with_uid=False):
#        if len(self.dict) == 0:
#            return iter(())
#        else:
#            start0, end0 = fetch_arghandler(start0, end0, self.fetch_start0, self.fetch_end0)
#            if start0 >= end0:
#                return iter(())
#            else:
#                yielder = self.get_yielder(with_uid)
#
#                for uid, read in self.dict.items():
#                    if read.reference_start >= end0:
#                        break
#                    if read.reference_end > start0:
#                        yield yielder(uid, read)
#
#    def _fetch_forward(self, fetch_start0, fetch_end0, with_uid=False):
#        yielder = self.get_yielder(with_uid)
#            
#        is_area1 = False
#        is_area2 = False
#        for uid, read in self.dict.items():
#            # pre-area1
#            if not is_area1:
#                # check if within area1
#                if read.reference_start > fetch_start0 - MAX_READ_LENGTH:
#                    is_area1 = True
#                else:
#                    # job pre-area1 (null)
#                    # continue
#                    continue
#            # area1
#            if not is_area2:
#                # check if within area2
#                if read.reference_start >= fetch_start0:
#                    is_area2 = True
#                else:
#                    # job area1
#                    if read.reference_end > fetch_start0:
#                        yield yielder(uid, read)
#                    # continue
#                    continue
#            # check if beyond area2
#            if read.reference_start >= fetch_end0:
#                break
#            # job area2
#            yield yielder(uid, read)
#                
#    def _fetch_reverse(self, fetch_start0, fetch_end0, with_uid=False):
#        yielder = self.get_yielder(with_uid)
#            
#        is_area1 = False
#        is_area2 = False
#        for uid, read in reversed(self.dict.items()):
#        #for uid, read in zip(reversed(self.uids), reversed(self.list)):
#            # post-area2
#            if not is_area2:
#                # check if within area2
#                if read.reference_start < fetch_end0:
#                    is_area2 = True
#                else:
#                    # job post-area2 (null)
#                    # continue
#                    continue
#            # area2
#            if not is_area1:
#                # check if within area1
#                if read.reference_start < fetch_start0:
#                    is_area1 = True
#                else:
#                    # job area2
#                    yield yielder(uid, read)
#                    # continue
#                    continue
#            # check if within pre-area1
#            if read.reference_start <= fetch_start0 - MAX_READ_LENGTH:
#                break
#            # job area1
#            if read.reference_end > fetch_start0:
#                yield yielder(uid, read)
#                                    
#    def fetch_choosing_side(self, start0=None, end0=None, with_uid=False):
#        if len(self.dict) == 0:
#            return iter(())
#        else:
#            start0, end0 = fetch_arghandler(start0, end0, self.fetch_start0, self.fetch_end0)
#            if start0 >= end0:
#                return iter(())
#            else:
#                # return self._fetch_forward(start0, end0, with_uid=with_uid)                
#                dist_from_start = start0 - self.fetch_start0
#                dist_from_end = self.fetch_end0 - end0
#
#                if dist_from_start <= dist_from_end:
#                    return self._fetch_forward(
#                        start0, end0, with_uid=with_uid
#                    )
#                else:
#                    return reversed(
#                        tuple(
#                            self._fetch_reverse(
#                                start0, end0, with_uid=with_uid
#                            )
#                        )
#                    )
#
#
#class FetchedReadsSinglechrom:
#    def __init__(self, chrom, bam, readfilter=None):
#        self.chrom = chrom
#        self.bam = bam
#
#        if readfilter is None:
#            self.readfilter = readhandler.readfilter_pileup
#        else:
#            self.readfilter = readfilter
#
#        self.clusters = list()
#
#    def __repr__(self):
#        return f'<FetchedReadsSinglechrom object (chrom={self.chrom}, clusters={self.clusters})>'
#
#    @classmethod
#    def from_fetch(cls, bam, chrom, start0, end0, readfilter=None):
#        result = cls(chrom, bam, readfilter)
#        result.clusters.append(FetchedReadsCluster.from_fetch(result.bam, result.chrom, start0, end0, result.readfilter))
#
#        return result
#
#    @property
#    def gr(self):
#        starts = list()
#        ends = list()
#        ids = list()
#        for x in self.clusters:
#            starts.append(x.fetch_start0)
#            ends.append(x.fetch_end0)
#            ids.append(x.get_id())
#        chroms = [self.chrom] * len(self.clusters)
#        return pr.from_dict(
#            {'Chromosome': chroms, 'Start': starts, 'End': ends, 'Id': ids}
#        )
#
##        return pr.PyRanges(
##            chromosomes=([self.chrom] * len(self.clusters)),
##            starts=[x.fetch_start0 for x in self.clusters],
##            ends=[x.fetch_end0 for x in self.clusters],
##        )
#
#    def get_clusters_by_ids(self, ids):
#        result = dict()
#        for x in self.clusters:
#            id = x.get_id()
#            if id in ids:
#                result[id] = x
#        return result
#
#    def get_read(self, read_uid):
#        found_read = False
#        for frcluster in self.clusters:
#            try:
#                read = frcluster.dict[read_uid]
#            except KeyError:
#                continue
#            else:
#                found_read = True
#                break
#
#        if not found_read:
#            raise Exception(f'Given ReadUID is not present.')
#
#        return read
#
#    def iter_reads(self):
#        return itertools.chain.from_iterable(x.dict.items() for x in self.clusters)
#
#    def sort_clusters(self):
#        self.clusters.sort(key=(lambda x: x.fetch_start0))
#
#    def merge(self):
#        new_clusters = list()
#        self.sort_clusters()
#
#        def unit_func(cluster1, cluster2):
#            if cluster1.fetch_end0 >= cluster2.fetch_start0:
#                return cluster1.merge(cluster2)
#            else:
#                new_clusters.append(cluster1)
#                return cluster2
#
#        final = functools.reduce(unit_func, self.clusters)
#        new_clusters.append(final)
#        self.clusters = new_clusters
#
#    def add_reads(self, start0, end0):
#        if len(self.clusters) == 0:
#            self.clusters.append(FetchedReadsCluster.from_fetch(self.bam, self.chrom, start0, end0, self.readfilter))
#        else:
#            self_gr = self.gr
#            input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
#            subtracted_gr = input_gr.subtract(self_gr)
#                # When there is nothing left after subtraction, "subtracted_gr" becomes an empty PyRanges. 
#            nearest_gr = subtracted_gr.nearest(self_gr)
#                # "nearest_gr" becomes also an empty PyRanges. 
#                # Iteration with .df.iterrows() results in an empty iteration.
#            # get adjacent clusters
#            adj_frcluster_ids = list()
#            for idx, row in nearest_gr.df.iterrows():
#                if abs(row['Distance']) == 1:
#                    adj_frcluster_ids.append(row['Id'])
#            adj_frclusters = self.get_clusters_by_ids(adj_frcluster_ids)
#            # do read addition
#            for idx, row in nearest_gr.df.iterrows():
#                if abs(row['Distance']) == 1:
#                    frcluster = adj_frclusters[row['Id']]
#                    added_region_width = row['End'] - row['Start']
#                    if row['Start'] == row['End_b']:
#                        frcluster.extend_rightward(row['End'] - row['Start'])
#                    elif row['End'] == row['Start_b']:
#                        frcluster.extend_leftward(row['End'] - row['Start'])
#                else:
#                    self.clusters.append(FetchedReadsCluster.from_fetch(self.bam, self.chrom, row['Start'], row['End'], self.readfilter))
#
#            if len(adj_frclusters) > 0:
#                self.merge()
#
#    def fetch(self, start0, end0, with_uid=False):
#        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
#        joined_gr = input_gr.join(self.gr, how=None)
#
#        valid_frcluster_ids = list()
#        for idx, row in joined_gr.df.iterrows():
#            if row['Start_b'] <= row['Start'] and row['End_b'] >= row['End']:
#                valid_frcluster_ids.append(row['Id'])
#
#        if len(valid_frcluster_ids) == 0:
#            raise Exception(f'Input fetch range is not fully contained within database.')
#        if len(valid_frcluster_ids) > 1:
#            raise Exception(f'There are more than one FetchedReadsCluster objects containing input fetch range.')
#
#        frcluster_id = valid_frcluster_ids[0]
#        frcluster = self.get_clusters_by_ids([frcluster_id])[frcluster_id]
#        return frcluster.fetch(start0, end0, with_uid=with_uid)
#
#    def check_fully_includes(self, start0, end0):
#        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
#        intersect_gr = input_gr.intersect(self.gr, how='containment')
#        return not intersect_gr.empty
#
#    def choose_cluster(self, start0, end0):
#        self_gr = self.gr
#        input_gr = pr.PyRanges(chromosomes=[self.chrom], starts=[start0], ends=[end0])
#        joined_gr = input_gr.join(self_gr, how=None)
#        if joined_gr.df.shape[0] == 0:
#            return None
#        else:
#            overlapping_cluster_ids = {row['Id'] for idx, row in joined_gr.df.iterrows()}
#            overlapping_clusters = list()
#            for idx, row in self_gr.df.iterrows():
#                if row['Id'] in overlapping_cluster_ids:
#                    overlapping_clusters.append(self.clusters[idx])
#
#            if len(overlapping_clusters) == 1:
#                return overlapping_clusters[0]
#            else:
#                raise Exception(f'More than one clusters overlap with input range. Overlapping clusters: {overlapping_clusters}')
#        
#
#class FetchedReads:
#    def __init__(self, bam, readfilter=None):
#        self.bam = bam
#
#        if readfilter is None:
#            self.readfilter = readhandler.readfilter_pileup
#        else:
#            self.readfilter = readfilter
#
#        self.subdata = dict()
#
#    def __repr__(self):
#        buffer = list()
#        buffer.append(f'<FetchedReads object, which consists of:')
#        for chrom, frsinglechrom in self.subdata.items():
#            buffer.append(f'\t{frsinglechrom}')
#        buffer.append(f'>')
#        return '\n'.join(buffer)
#
#    @classmethod
#    def from_fetch(cls, bam, chrom, start0, end0, readfilter=None):
#        result = cls(bam, readfilter)
#        result.subdata[chrom] = FetchedReadsSinglechrom.from_fetch(result.bam, chrom, start0, end0, result.readfilter)
#
#        return result
#
#    def get_read(self, read_uid):
#        found_read = False
#        for frcluster in itertools.chain.from_iterable(
#            (cluster for cluster in frsinglechr.clusters)
#            for frsinglechr in self.subdata.values()
#        ):
#            try:
#                read = frcluster.dict[read_uid]
#            except KeyError:
#                continue
#            else:
#                found_read = True
#                break
#
#        if not found_read:
#            raise Exception(f'Given ReadUID is not present.')
#
#        return read
#
#    def iter_reads(self, chrom=None):
#        if chrom is None:
#            return itertools.chain.from_iterable(x.iter_reads() for x in self.subdata.values())
#        else:
#            return self.subdata[chrom].iter_reads()
#
#    def add_reads(self, chrom, start0, end0):
#        if chrom not in self.subdata:
#            self.subdata[chrom] = FetchedReadsSinglechrom(chrom, self.bam, self.readfilter)
#        self.subdata[chrom].add_reads(start0, end0)
#
#    def fetch(self, chrom, start0, end0, with_uid=False):
#        if chrom not in self.subdata:
#            raise Exception(f'Input chrom is not included in the database.')
#        return self.subdata[chrom].fetch(start0, end0, with_uid=with_uid)
#
#    def check_fully_includes(self, chrom, start0, end0):
#        if chrom not in self.subdata:
#            return False
#        else:
#            return self.subdata[chrom].check_fully_includes(start0, end0)
#
#    def choose_cluster(self, chrom, start0, end0):
#        if chrom not in self.subdata:
#            return False
#        else:
#            return self.subdata[chrom].choose_cluster(start0, end0)


########################


class PileupBase:
    def _coord_arg_sanitycheck(self, start0, end0):
        #if start0 is not None:
        if start0 < self.start0:
            raise Exception('Input "start0" argument is out of pileup range.')
        #if end0 is not None:
        if end0 > self.end0:
            raise Exception('Input "end0" argument is out of pileup range.')

    def _coord_arg_sanitycheck_pos0(self, pos0):
        if pos0 not in self.df.columns:
            raise Exception('Input "pos0" argument is out of pileup range.')

    @property
    def start0(self):
        return self.df.columns[0]

    @property
    def end0(self):
        return self.df.columns[-1] + 1

    @property
    def range0(self):
        return range(self.start0, self.end0)

    def get_ref_seq(self, start0=None, end0=None):
        if start0 is None:
            start0 = self.start0
        if end0 is None:
            end0 = self.end0
        self._coord_arg_sanitycheck(start0, end0)

        key = (start0, end0)
        if key not in self._ref_seq_cache:
            self._ref_seq_cache[key] = self.fasta.fetch(self.chrom, key[0], key[1])
        return self._ref_seq_cache[key]

#    @staticmethod
#    def get_max_vaf_from_pileupcol(col, with_counts=False, empty_value=EMPTY_VALUE):
#        counts = collections.Counter(x for x in col if x != empty_value)
#        total = sum(counts.values())
#        if total == 0:
#            raise Exception(f"Pileup column allele count sum is 0.")
#
#        max_allele, max_count = max(counts.items(), key=(lambda x: x[1]))
#        max_vaf = max_count / total
#
#        if with_counts:
#            return max_allele, max_vaf, counts
#        else:
#            return max_allele, max_vaf
#
#    def get_max_vaf_colindex(self, col_idx, with_counts=False):
#        col = self.df.iloc[:, col_idx]
#        return self.get_max_vaf_from_pileupcol(col, with_counts=with_counts)
#
#    def get_max_vaf_colpos0(self, col_pos0, with_counts=False):
#        col = self.df.loc[:, col_pos0]
#        return self.get_max_vaf_from_pileupcol(col, with_counts=with_counts)

    def get_allele_counter(self, pos0):
        #self._coord_arg_sanitycheck_pos0(pos0)
        col = self.df.loc[:, pos0]
        return collections.Counter(x for x in col if x != EMPTY_VALUE)

    def get_allele_portions(self, pos0):
        counter = self.get_allele_counter(pos0)
        counter_sum = sum(counter.values())
        portions = dict()
        for key, val in counter.items():
            portions[key] = val / counter_sum
        return portions

    # active info related ones #
    def iter_active_info_old_deprecated(self, start0, end0, reverse=False):
        self._coord_arg_sanitycheck(start0, end0)

        ref_seq = self.get_ref_seq(start0, end0)
        pos0_range = range(start0, end0)
        if reverse:
            ref_seq = ref_seq[::-1]
            pos0_range = pos0_range[::-1]

        for ref_base, pos0 in zip(ref_seq, pos0_range):
            max_allele, max_vaf = self.get_max_vaf_colpos0(pos0, with_counts=False)
            if max_vaf >= self.active_threshold:
                is_active = not (max_allele == ref_base)
            else:
                is_active = True
            yield is_active

    def iter_active_info(self, start0, end0, reverse=False):
        self._coord_arg_sanitycheck(start0, end0)

        ref_seq = self.get_ref_seq(start0, end0)
        pos0_range = range(start0, end0)
        if reverse:
            ref_seq = ref_seq[::-1]
            pos0_range = pos0_range[::-1]

        for ref_base, pos0 in zip(ref_seq, pos0_range):
            counter = self.get_allele_counter(pos0)
            non_ref_portion = 1 - (counter[ref_base] / sum(counter.values()))
            is_active = (non_ref_portion >= self.active_threshold)
            yield is_active 

    def _set_active_info(self, start0=None, end0=None):
        if start0 is None:
            start0 = self.start0
        if end0 is None:
            end0 = self.end0

        self._coord_arg_sanitycheck(start0, end0)
        if not hasattr(self, '_active_info'):
            setattr(self, '_active_info', list())

        active_info_idx_start = start0 - self.start0
        active_info_idx_end = end0 - self.start0

        self._active_info[
            active_info_idx_start:active_info_idx_end
        ] = self.iter_active_info(start0, end0, reverse=False)

    def get_active_info(self, start0=None, end0=None):
        if start0 is None:
            start0 = self.start0
        if end0 is None:
            end0 = self.end0

        self._coord_arg_sanitycheck(start0, end0)

        active_info_idx_start = start0 - self.start0
        active_info_idx_end = end0 - self.start0

        return pd.Series(
            self._active_info[active_info_idx_start:active_info_idx_end], 
            index=self.df.columns[active_info_idx_start:active_info_idx_end],
        )

    def get_active_positions(self):
        return tuple(
            itertools.compress(self.df.columns, self._active_info)
        )

    def check_inactive_margin_right(self, length):
        return not any(self._active_info[-length:])

    def check_inactive_margin_left(self, length):
        return not any(self._active_info[:length])

    # row specs related ones
    @staticmethod
    def row_spec_matcher(query, target):
        if query['left_filled']:
            if query['right_filled']:
                if target['left_filled'] and target['right_filled']:
                    return query['seq'] in target['seq']
                else:
                    return False
            else:
                if target['left_filled']:
                    return query['seq'] == target['seq'][:len(query['seq'])]
                else:
                    return False
        else:
            if query['right_filled']:
                if target['right_filled']:
                    return query['seq'] == target['seq'][-len(query['seq']):]
                else:
                    return False
            else:
                return query['seq'] in target['seq']

    def set_row_spec_groups(self):
        """row_spec's are grouped by whether one is a substring of another."""
        self.row_spec_groups = collections.OrderedDict()
        for query_rowid, query in sorted(
            self.row_specs.items(), 
            key=(lambda x: x[1]['match_length']), 
            reverse=True,
        ):
            superseq_candidates = list()
            for target_rowid in self.row_spec_groups.keys():
                target = self.row_specs[target_rowid]
                if self.row_spec_matcher(query, target):
                    superseq_candidates.append(target_rowid)
                if len(superseq_candidates) >= 2:
                    break
                    
            if len(superseq_candidates) == 0:
                self.row_spec_groups[query_rowid] = {
                    'subseq_rowids': set(),
                    'subseq_hits': 0,
                }
                self.row_spec_groups[query_rowid]['subseq_rowids'].add(query_rowid)
                self.row_spec_groups[query_rowid]['subseq_hits'] += 1
            elif len(superseq_candidates) == 1:
                superseq_rowid = superseq_candidates[0]
                self.row_spec_groups[superseq_rowid]['subseq_rowids'].add(query_rowid)
                self.row_spec_groups[superseq_rowid]['subseq_hits'] += 1
            else:
                for superseq_rowid in superseq_candidates:
                    self.row_spec_groups[superseq_rowid]['subseq_rowids'].add(query_rowid)

    # for debugging
    def show_row_spec_alignments(self):
        subseq_hits_sum = sum(x['subseq_hits'] for x in self.row_spec_groups.values())
        for superseq_id, groupinfo in sorted(
            self.row_spec_groups.items(),
            key=(lambda x: x[1]['subseq_hits']),
            reverse=True,
        ):
            superseq_row_spec = self.row_specs[superseq_id]
            lstrip_query_gaps = not superseq_row_spec["left_filled"]
            rstrip_query_gaps = not superseq_row_spec["right_filled"]

            superseq_aln = self._alignment_cache[superseq_id]
            superseq_vcfspec_list = alignhandler.alignment_to_vcfspec(
                superseq_aln, 
                self.start0, 
                self.chrom, 
                self.fasta,
                lstrip_query_gaps=lstrip_query_gaps,
                rstrip_query_gaps=rstrip_query_gaps,
            )

            print('subseq hits:', groupinfo['subseq_hits'])
            print('subseq hits portion:', groupinfo['subseq_hits'] / subseq_hits_sum)
            print(superseq_row_spec)
            print(superseq_vcfspec_list)
            print(superseq_aln)
            print()


class Pileup(PileupBase):
    pat_insclip = re.compile("(\(.+\))?([^()]+)(\(.+\))?")
    pat_parenthesis = re.compile("[()]")
    row_spec_exluded_vals = {DEL_VALUE, EMPTY_VALUE}

    def __init__(self, df, chrom, fasta, fetchedreads, active_threshold=None):
        self.df = df
        self.chrom = chrom
        self.fasta = fasta
        self.fetchedreads = fetchedreads
        self.bam = self.fetchedreads.bam

        if active_threshold is None:
            self.active_threshold = librealign.DEFAULT_ACTIVE_THRESHOLD
        else:
            self.active_threshold = active_threshold

        self._ref_seq_cache = dict()
        self._alignment_cache = dict()

        self._set_active_info()

    def pick_frcluster(self):
        return self.fetchedreads.choose_cluster(self.chrom, self.start0, self.end0)

    def subset(self, start0, end0, inplace=False):
        self._coord_arg_sanitycheck(start0, end0)
        # subset dataframe
        new_df = self.df.loc[:, start0:(end0 - 1)]
        # remove out-of-range rows
        row_within_range = list()
        for row_id, row in new_df.iterrows():
            read = self.get_read(row_id)
            if read.reference_end <= start0 or read.reference_start >= end0:
                row_within_range.append(False)
            else:
                row_within_range.append(True)
        new_df = new_df.loc[row_within_range, :]
        # active_info
        active_info_idx_start = start0 - self.start0
        active_info_idx_end = end0 - self.start0
        new_active_info = self._active_info[active_info_idx_start:active_info_idx_end]

        # result
        if inplace:
            self.df = new_df
            self._active_info = new_active_info
        else:
            result = self.__class__(
                df=new_df, chrom=self.chrom, fasta=self.fasta,
                fetchedreads=self.fetchedreads, active_threshold=self.active_threshold,
            )
            result._active_info = new_active_info
            result._ref_seq_cache = self._ref_seq_cache
            return result

    def get_read(self, row_id):
        return self.fetchedreads.get_read(row_id)

    ### extend ###  
    def extend_rightward(self, width):
        # keep original starts and ends
        border_col_idx_left = self.df.shape[1] - 1
        # make new pileup
        new_pileup = self._extend_helper_make_new_pileup(
            chrom=self.chrom,
            start0=self.end0,
            end0=(self.end0 + width),
            fetchedreads=self.fetchedreads,
        )
        # join
        self.df = self.df.join(new_pileup.df, how="outer")
        self.df[self.df.isnull()] = EMPTY_VALUE  # turn NaN into EMPTY_VALUE
        self._active_info.extend(new_pileup._active_info)
        # handle insclips facing each other
        self._extend_helper_handle_facing_insclip(border_col_idx_left)

    def extend_leftward(self, width):
        # keep original starts and ends
        border_col_idx_left = width - 1
        # make new pileup
        new_pileup = self._extend_helper_make_new_pileup(
            chrom=self.chrom,
            start0=(self.start0 - width),
            end0=self.start0,
            fetchedreads=self.fetchedreads,
        )
        # join
        self.df = new_pileup.df.join(self.df, how="outer")
        self.df[self.df.isnull()] = EMPTY_VALUE  # turn NaN into EMPTY_VALUE
        self._active_info = new_pileup._active_info + self._active_info
        # handle insclips facing each other
        self._extend_helper_handle_facing_insclip(border_col_idx_left)

    def _extend_helper_handle_facing_insclip(self, border_col_idx_left):
        border_col_idx_right = border_col_idx_left + 1
        border_col_left = self.df.iloc[:, border_col_idx_left]
        border_col_right = self.df.iloc[:, border_col_idx_right]
        for row_idx, (val_left, val_right) in enumerate(
            zip(border_col_left, border_col_right)
        ):
            if val_right.startswith("(") and val_left.endswith(")"):
                mat_left = self.__class__.pat_insclip.fullmatch(val_left)
                mat_right = self.__class__.pat_insclip.fullmatch(val_right)
                if mat_left.group(3) != mat_right.group(1):
                    raise Exception(
                        f"Insclip seqs of adjacent entries are different.\n{self.df}"
                    )
                self.df.iloc[
                    row_idx, border_col_idx_right
                ] = self.__class__.pat_insclip.sub("\\2\\3", "", val_right)

    def _extend_helper_make_new_pileup(self, chrom, start0, end0, fetchedreads):
        return get_pileup(
            chrom, 
            start0, 
            end0,
            fetchedreads=fetchedreads,
            fasta=self.fasta,
            active_threshold=self.active_threshold,
            truncate=True,
            as_array=False,
            return_range=False,
        )

    ################################
    # realignment-related features #
    ################################

    # secure_inactive_padding
    def secure_inactive_padding_rightward(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY):
        librealign.secure_inactive_padding_rightward(
            pileup=self,
            inactive_padding=inactive_padding, 
            extend_pileup_by=extend_pileup_by, 
            inplace=True,
        )

    def secure_inactive_padding_leftward(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY):
        librealign.secure_inactive_padding_leftward(
            pileup=self, 
            inactive_padding=inactive_padding, 
            extend_pileup_by=extend_pileup_by, 
            inplace=True,
        )

    def secure_inactive_padding(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY):
        self.secure_inactive_padding_rightward(inactive_padding, extend_pileup_by=extend_pileup_by)
        self.secure_inactive_padding_leftward(inactive_padding, extend_pileup_by=extend_pileup_by)

    ### row specs ###
    def get_row_spec(self, row_id):
        row = self.df.loc[row_id, :]
        return {
            "seq": "".join(
                self.__class__.pat_parenthesis.sub("", x) 
                for x in row if x not in self.__class__.row_spec_exluded_vals
            ),
            "match_length": sum(x != EMPTY_VALUE for x in row),
            "left_filled": (row.iloc[0] != EMPTY_VALUE),
            "right_filled": (row.iloc[-1] != EMPTY_VALUE),
            "id": row.name,
        }

    def set_row_specs(self):
        self.row_specs = dict()
        for row_id, row in self.df.iterrows():
            self.row_specs[row_id] = self.get_row_spec(row_id)

    # others
    def save_alignments(self, superseq_rowid_list, aligner=None):
        if aligner is None:
            aligner = librealign.ALIGNER_MAIN
        ref_seq = self.get_ref_seq()
        ref_seq_reversed = ref_seq[::-1]

        for row_id in superseq_rowid_list:
            self._alignment_cache[row_id] = librealign.get_alignment_from_rowid(row_id, self, ref_seq, ref_seq_reversed, aligner)

    def augment_range(self):
        active_region_pileup.set_row_specs()
        active_region_pileup.set_row_spec_groups()
        vcfspecs = active_region_pileup.get_vcfspecs()

    def get_realigned_reads(self, aligner=None):
        """Returns:
            dict (keys ReadUID, values pysam.AlignedSegment)
        """
        if aligner is None:
            aligner = librealign.ALIGNER_MAIN
        ref_seq = self.get_ref_seq()
        ref_seq_reversed = ref_seq[::-1]
        return librealign.get_realigned_reads(self, self.range0, ref_seq, ref_seq_reversed, aligner)

    def get_vcfspecs(self, allele_portion_threshold=None):
        if allele_portion_threshold is None:
            allele_portion_threshold = librealign.DEFAULT_VCFSPEC_EXTRACTION_THRESHOLD
        return librealign.get_vcfspecs_from_pileup(self, allele_portion_threshold=allele_portion_threshold)


class MultisamplePileup(PileupBase):
    def __init__(self, pileup_dict):
        """Args:
            pileup_dict: keys - sampleid; values - Pileup object
        """
        self.pileup_dict = pileup_dict

        #self.active_threshold = self._get_active_threshold()
        self._alignment_cache = dict()

        first_pileup = next(iter(self.pileup_dict.values()))
        self.chrom = first_pileup.chrom
        self.fasta = first_pileup.fasta
        self._ref_seq_cache = first_pileup._ref_seq_cache

    def _get_active_threshold(self):
        depth_sum = sum(x.df.shape[0] for x in self.pileup_dict.values())
        corrected_active_thresholds = (
            x.active_threshold * (x.df.shape[0] / depth_sum)
            for x in self.pileup_dict.values()
        )
        return min(corrected_active_thresholds)

    def _get_active_threshold_simple(self):
        return sum(x.active_threshold for x in self.pileup_dict.values()) / (len(self.pileup_dict) ** 2)

    def set_df(self):
        assert len(set(x.range0 for x in self.pileup_dict.values())) == 1, f'Genomic ranges of pileup objects are different.'
        self.df = pd.concat(
            {key: val.df for key, val in self.pileup_dict.items()},
            names=['SampleID', 'ReadUID'],
        )
        self._set_active_info()

    def get_read(self, row_id):
        sampleid, read_uid = row_id
        return self.pileup_dict[sampleid].get_read(read_uid)

    ### extend ###
    def extend_rightward(self, width, extend_fetchedreads_by=DEFAULT_EXTEND_FETCHEDREADS_BY):
        for pileup in self.pileup_dict.values():
            pileup.extend_rightward(width, extend_fetchedreads_by=extend_fetchedreads_by)

    def extend_leftward(self, width, extend_fetchedreads_by=DEFAULT_EXTEND_FETCHEDREADS_BY):
        for pileup in self.pileup_dict.values():
            pileup.extend_leftward(width, extend_fetchedreads_by=extend_fetchedreads_by)

    # equalize
    def equalize_margins_leftward(self):
        max_range_start0 = min(x.start0 for x in self.pileup_dict.values())
        for pileup in self.pileup_dict.values():
            if pileup.start0 > max_range_start0:
                pileup.extend_leftward(pileup.start0 - max_range_start0)

    def equalize_margins_rightward(self):
        max_range_end0 = max(x.end0 for x in self.pileup_dict.values())
        for pileup in self.pileup_dict.values():
            if pileup.end0 < max_range_end0:
                pileup.extend_rightward(max_range_end0 - pileup.end0)

    def equalize_margins(self):
        self.equalize_margins_leftward()
        self.equalize_margins_rightward()

    # secure inactive padding
    def secure_inactive_padding_leftward(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY, extend_fetchedreads_by=DEFAULT_EXTEND_FETCHEDREADS_BY):
        while True:
            self.equalize_margins_leftward()
            # check inactive margins
            margin_insufficient = list()
            for sampleid, pileup in self.pileup_dict.items():
                if not pileup.check_inactive_margin_left(inactive_padding):
                    margin_insufficient.append(sampleid)
            # check whether to break
            if len(margin_insufficient) == 0:
                break
            # secure inactive margins
            for sampleid in margin_insufficient:
                self.pileup_dict[sampleid].secure_inactive_padding_leftward(inactive_padding, extend_pileup_by=extend_pileup_by, extend_fetchedreads_by=extend_fetchedreads_by)

    def secure_inactive_padding_rightward(self, inactive_padding, extend_pileup_by=DEFAULT_EXTEND_PILEUP_BY, extend_fetchedreads_by=DEFAULT_EXTEND_FETCHEDREADS_BY):
        while True:
            self.equalize_margins_rightward()
            # check inactive margins
            margin_insufficient = list()
            for sampleid, pileup in self.pileup_dict.items():
                if not pileup.check_inactive_margin_right(inactive_padding):
                    margin_insufficient.append(sampleid)
            # check whether to break
            if len(margin_insufficient) == 0:
                break
            # secure inactive margins
            for sampleid in margin_insufficient:
                self.pileup_dict[sampleid].secure_inactive_padding_rightward(inactive_padding, extend_pileup_by=extend_pileup_by, extend_fetchedreads_by=extend_fetchedreads_by)

    ### row specs ###
    def set_row_specs(self):
        for pileup in self.pileup_dict.values():
            pileup.set_row_specs()

        self.row_specs = dict()
        for idx, row in self.df.iterrows():
            sampleid, read_uid = idx
            self.row_specs[(sampleid, read_uid)] = self.pileup_dict[sampleid].row_specs[read_uid]

    def divide_row_spec_groups(self):
        # initialize
        row_spec_groups_by_sample = dict()
        superseq_rowids_sorted = [
            superseq_rowid for superseq_rowid, groupinfo in sorted(
                self.row_spec_groups.items(),
                key=(lambda x: x[1]['subseq_hits']),
                reverse=True,
            )
        ]
        for sampleid, pileup in self.pileup_dict.items():
            row_spec_groups_by_sample[sampleid] = collections.OrderedDict()
            for superseq_rowid in superseq_rowids_sorted:
                row_spec_groups_by_sample[sampleid][superseq_rowid] = {
                    #'subseq_rowids': set(),
                    'subseq_hits': 0,
                }
        # fill in values
        for superseq_rowid, groupinfo in self.row_spec_groups.items():
            for subseq_rowid in groupinfo['subseq_rowids']:
                subseq_sampleid, subseq_readuid = subseq_rowid
                #row_spec_groups_by_sample[subseq_sampleid][superseq_rowid]['subseq_rowids'].add(subseq_readuid)
                row_spec_groups_by_sample[subseq_sampleid][superseq_rowid]['subseq_hits'] += 1
        # assign to each single sample pileup
        for sampleid, groups in row_spec_groups_by_sample.items():
            self.pileup_dict[sampleid].row_spec_groups = groups

    # get realigned reads #
    def get_realigned_reads(self, aligner=None):
        """Returns:
            dict (keys sampleid, values dict (keys ReadUID, values pysam.AlignedSegment))
        """
        if aligner is None:
            aligner = librealign.ALIGNER_MAIN
        ref_seq = self.get_ref_seq()
        ref_seq_reversed = ref_seq[::-1]

        realigned_reads_allsamples = librealign.get_realigned_reads(self, self.range0, ref_seq, ref_seq_reversed, aligner)
        realigned_reads_multisample = dict()
        for row_id, read in realigned_reads_allsamples.items():
            sampleid, readuid = row_id
            realigned_reads_multisample.setdefault(sampleid, dict())
            realigned_reads_multisample[sampleid][readuid] = read

        return realigned_reads_multisample

    def get_vcfspecs(self, allele_portion_threshold=None):
        """Must be run after 'divide_row_spec_groups' and 'get_realigned_reads' 
            methods have been run.
            'divide_row_spec_groups' is needed for setting 'row_spec_groups' for
                each singlesample Pileup.
            'get_realigned_reads' is needed for initiating '_alignment_cache' attribute.
        """
        if allele_portion_threshold is None:
            allele_portion_threshold = librealign.DEFAULT_VCFSPEC_EXTRACTION_THRESHOLD
        vcfspecs_allsamples = librealign.get_vcfspecs_multisample_pileup(self, allele_portion_threshold)
        return vcfspecs_allsamples


@common.get_deco_num_set(('bam', 'fetchedreads'), 1)
def get_pileup(
    chrom,
    start0,
    end0,
    bam=None,
    fetchedreads=None,
    fasta=None,
    active_threshold=None,
    truncate=True,
    as_array=False,
    return_range=False,
    del_value=DEL_VALUE,
    empty_value=EMPTY_VALUE,
):
    def sanity_check(chrom, start0, end0, bam, fetchedreads, fasta, as_array):
        if (not as_array) and (fasta is None):
            raise Exception(f'If "as_array" is False, "fasta" must be set.')

#    def prepare_fetchedreads(bam, fetchedreads):
#        if fetchedreads is None:
#            fetchedreads = fetchcache.FetchedReads(bam, readfilter=readhandler.readfilter_pileup)
#        return fetchedreads

    def make_readlist(fetchedreads, chrom, start0, end0):
        readlist = list()
        start_list = list()
        end_list = list()
        uid_list = list()
        for uid, read in fetchedreads.fetch(chrom, start0, end0, with_uid=True):
            readlist.append(read)
            start_list.append(read.reference_start)
            end_list.append(read.reference_end)
            uid_list.append(uid)
        tmp_pileup_range = range(min(start_list), max(end_list))

        return readlist, tmp_pileup_range, uid_list

    def raise_cigarpattern_error(read):
        raise Exception(f"Unexpected cigar pattern:\n{read.to_string()}")

    def cigar_sanitycheck(read):
        """Assumes M.* or [IS]M.*"""
        error = False
        first_cigarop = read.cigartuples[0][0]
        if first_cigarop != 0:
            if first_cigarop not in (1, 4):
                error = True
            else:
                if read.cigartuples[1][0] != 0:
                    error = True
        if error:
            raise_cigarpattern_error(read)

    def write_read_to_array_row(read, arr, arr_row_idx, initial_arr_col_idx, del_value):
        def handle_leading_insclip(read):
            leading_insclip_len = 0
            cigartup_idx = 0
            for cigarop, cigarlen in alignhandler.iter_leading_queryonly(read.cigartuples):
                leading_insclip_len += cigarlen
                cigartup_idx += 1

            if leading_insclip_len == 0:
                leading_insseq = None
            else:
                leading_insseq = read.query_sequence[:leading_insclip_len]
            query_idx = leading_insclip_len

            return leading_insseq, cigartup_idx, query_idx

        def handle_queryonly_only_case(read, arr, arr_col_idx, arr_row_idx):
            arr[arr_row_idx, arr_col_idx] = f"({read.query_sequence})"

        def handle_targetonly_after_match(
            read,
            arr,
            query_idx,
            arr_col_idx,
            arr_row_idx,
            trailing_nonM_cigarlen_sum,
            del_value,
        ):
            arr[arr_row_idx, arr_col_idx] = read.query_sequence[
                query_idx
            ]  # query_idx indicates the last base of M
            arr[
                arr_row_idx,
                (arr_col_idx + 1) : (arr_col_idx + 1 + trailing_nonM_cigarlen_sum),
            ] = del_value

        def handle_queryonly_after_match(
            read,
            arr,
            query_idx,
            arr_col_idx,
            arr_row_idx,
            trailing_nonM_cigarlen_sum,
        ):
            last_match_base = read.query_sequence[query_idx]
            insseq = read.query_sequence[
                (query_idx + 1) : (query_idx + 1 + trailing_nonM_cigarlen_sum)
            ]
            arr[arr_row_idx, arr_col_idx] = f"{last_match_base}({insseq})"

        def handle_nonlast_match(
            read,
            arr,
            cigarlen,
            query_idx,
            arr_col_idx,
            arr_row_idx,
            cigartup_idx,
        ):
            # adds all match seqs except the last one
            # query_idx and arr_col_idx updated
            if cigarlen > 1:
                sl_query = slice(query_idx, query_idx + cigarlen - 1)
                sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen - 1)
                arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])
                query_idx += cigarlen - 1
                arr_col_idx += cigarlen - 1
            # get all subsequent non-M cigar units
            # cigartup_idx updated
            trailing_nonMs = list()
            while True:
                cigartup_idx += 1
                if cigartup_idx >= len(read.cigartuples):
                    break
                else:
                    next_cigartup = read.cigartuples[cigartup_idx]
                    if next_cigartup[0] != 0:
                        trailing_nonMs.append(next_cigartup)
                    else:
                        break
            # add trailing non-M cigar units
            # query_idx and arr_col_idx updated
            trailing_nonM_cigarops = set(x[0] for x in trailing_nonMs)
            trailing_nonM_cigarlen_sum = sum(x[1] for x in trailing_nonMs)
            if trailing_nonM_cigarops.issubset(alignhandler.CIGAROPS_TARGETONLY):  # D, N
                handle_targetonly_after_match(
                    read,
                    arr,
                    query_idx,
                    arr_col_idx,
                    arr_row_idx,
                    trailing_nonM_cigarlen_sum,
                    del_value,
                )
                query_idx += 1
                arr_col_idx += trailing_nonM_cigarlen_sum + 1
            elif trailing_nonM_cigarops.issubset(alignhandler.CIGAROPS_QUERYONLY):  # I, S
                handle_queryonly_after_match(
                    read,
                    arr,
                    query_idx,
                    arr_col_idx,
                    arr_row_idx,
                    trailing_nonM_cigarlen_sum,
                )
                query_idx += trailing_nonM_cigarlen_sum + 1
                arr_col_idx += 1
            else:
                raise Exception(
                    f"Consecutive Non-M cigar units are composed of both target-only and query-only ones:\n{read.to_string()}"
                )

            return cigartup_idx, query_idx, arr_col_idx

        def handle_last_match(
            read,
            arr,
            cigarlen,
            query_idx,
            arr_col_idx,
            arr_row_idx,
        ):
            # adds all match seqs to array
            sl_query = slice(query_idx, query_idx + cigarlen)
            sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen)
            arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])

        # main
        """1) First, leading query-only cigar units are collected.
        2) Then, cigartuple is iterated on. Initial cigartup_idx is at the 
            first cigar unit which is not a query-only one. This cigar unit
            is assumed to be M. (If not, raises an exception)
        3) What is done in a loop cycle:
            The cigar unit which cigartup_idx indicates (which should be M)
            and all subsequent non-M cigarop are treated at single loop cycle.
            The subsequent non-M cigar units are assumed to be composed of only
            query-only ones or target-only ones. (If not, raises an exception)
        """

        # set mutable arr_col_idx. initial one is kept separately.
        arr_col_idx = initial_arr_col_idx

        # handles leading query-only cigar units. cigartup_idx and query_idx are initialized
        leading_insseq, cigartup_idx, query_idx = handle_leading_insclip(read)

        if cigartup_idx == len(read.cigartuples):
            # cigartup is only composed of query-only operations (no M)
            arr[arr_row_idx, arr_col_idx] = "(" + read.query_sequence + ")"            
        else:

            
            if read.cigartuples[cigartup_idx][0] != 0:
                raise Exception(
                    f"The first cigar operation after stripping leading I and S is not M:\n{read.to_string()}"
                )

            # begins loop
            while True:
                if cigartup_idx > len(read.cigartuples) - 1:
                    raise Exception(
                        f'"cigartup_idx" became greater than "len(read.cigartuples) - 1" while looping.'
                    )

                cigarop, cigarlen = read.cigartuples[cigartup_idx]
                assert (
                    cigarop == 0
                ), f'The cigar unit indicated by "cigarop_idx" at the beginning of a loop cycle is not M.'

                if (
                    cigartup_idx == len(read.cigartuples) - 1
                ):  # current cigartup is the last one
                    handle_last_match(
                        read,
                        arr,
                        cigarlen,
                        query_idx,
                        arr_col_idx,
                        arr_row_idx,
                    )
                    
                    break
                else:
                    cigartup_idx, query_idx, arr_col_idx = handle_nonlast_match(
                        read,
                        arr,
                        cigarlen,
                        query_idx,
                        arr_col_idx,
                        arr_row_idx,
                        cigartup_idx,
                    )
                    if cigartup_idx == len(read.cigartuples):
                        break

            # add leading query-only seq
            if leading_insseq is not None:
                arr[arr_row_idx, initial_arr_col_idx] = (
                    f"({leading_insseq})" + arr[arr_row_idx, initial_arr_col_idx]
                )

    def truncate_array(arr, tmp_pileup_range, pileup_range):
        sl_start = tmp_pileup_range.index(pileup_range.start)
        if pileup_range.stop == tmp_pileup_range.stop:
            sl_end = None
        else:
            sl_end = tmp_pileup_range.index(pileup_range.stop)
        return arr[:, slice(sl_start, sl_end)]

    def make_pileup_from_array(
        chrom,
        arr,
        fetchedreads,
        uid_list,
        pileup_range,
        tmp_pileup_range,
        truncate,
        active_threshold,
        fasta,
    ):
        df_columns = pileup_range if truncate else tmp_pileup_range
        df = pd.DataFrame(arr, columns=list(df_columns), index=uid_list)
        pileup = Pileup(
            df=df,
            chrom=chrom,
            fasta=fasta,
            fetchedreads=fetchedreads,
            active_threshold=active_threshold,
        )
        return pileup

    # main
    # parameter handling
    sanity_check(chrom, start0, end0, bam, fetchedreads, fasta, as_array)
    if active_threshold is None:
        active_threshold = librealign.DEFAULT_ACTIVE_THRESHOLD
    if fetchedreads is None:
        fetchedreads = fetchcache.FetchedReads(bam, readfilter=readhandler.readfilter_pileup)
    # prepare pileup_range
    pileup_range = range(start0, end0)
    #pileup_range, fetchedreads = set_params(chrom, start0, end0, bam, fetchedreads)
    readlist, tmp_pileup_range, uid_list = make_readlist(fetchedreads, chrom, start0, end0)
    # create array
    arr = np.full(
        shape=(len(readlist), len(tmp_pileup_range)),
        fill_value=empty_value,
        dtype=object,
    )
    for arr_row_idx, read in enumerate(readlist):
        initial_arr_col_idx = read.reference_start - tmp_pileup_range.start
        try:
            write_read_to_array_row(read, arr, arr_row_idx, initial_arr_col_idx, del_value)
        except Exception as exc:
            try:
                cigar_sanitycheck(read)
            except Exception as exc_cigarpattern:
                raise exc_cigarpattern from exc
            else:
                raise exc
    # truncate array
    if truncate:
        arr = truncate_array(arr, tmp_pileup_range, pileup_range)
    # prepare results
    if as_array:
        if return_range:
            return (arr, pileup_range)
        else:
            return arr
    else:
        pileup = make_pileup_from_array(
            chrom,
            arr,
            fetchedreads,
            uid_list,
            pileup_range,
            tmp_pileup_range,
            truncate,
            active_threshold,
            fasta,
        )
        if return_range:
            return (pileup, pileup_range)
        else:
            return pileup


def get_pileup_multisample(
    chrom,
    start0,
    end0,
    bam_dict,
    fasta,
    active_threshold_onesample=None,
    del_value=DEL_VALUE,
    empty_value=EMPTY_VALUE,
):
    # parameter handling
    if active_threshold_onesample is None:
        active_threshold_onesample = librealign.DEFAULT_ACTIVE_THRESHOLD
    # initialize pileup_dict
    pileup_dict = dict()
    for sampleid, bam in bam_dict.items():
        pileup_dict[sampleid] = get_pileup(
            chrom,
            start0,
            end0,
            bam=bam,
            fasta=fasta,
            active_threshold=active_threshold_onesample,
            truncate=True,
            as_array=False,
            return_range=False,
            del_value=del_value,
            empty_value=empty_value,
        )
    # create MultisamplePileup object and postprocess
    mspileup = MultisamplePileup(pileup_dict)
    mspileup.set_df()

    return mspileup



# def get_pileup(bam, chrom, start0, end0, as_array=False, truncate=True, with_reads=False):
#    """Returns:
#        A numpy.ndarray object if "as_array" argument is True
#        A Pileup object if "as_array" argument is False
#    """
#
#    read_dict = {
#        readhandler.get_uid(x): x
#        for x in readhandler.get_fetch(bam, chrom, start0, end0, readfilter=readhandler.readfilter_pileup)
#    }
#    readlist = tuple(iter(read_dict.values()))
#    pileup, pileup_range = get_pileup_from_reads(readlist, as_array=as_array, return_range=True)
#
#    if truncate:
#        if as_array:
#            start0_idx = pileup_range.index(start0)
#            end0_idx = pileup_range.index(end0)
#            pileup = pileup[:, start0_idx:end0_idx]
#        else:
#            pileup.df = pileup.df.loc[:, list(range(start0, end0))]
#
#    if with_reads:
#        return pileup, read_dict
#    else:
#        return pileup


# def get_pileup_from_reads(readlist, as_array=False, pileup_range=None, return_range=False, del_char='*'):
#    def raise_err(read):
#        raise Exception(f"Unexpected cigar pattern:\n{read.to_string()}")
#
#    def cigar_sanitycheck(read):
#        """Assumes M.* or [IS]M.*"""
#        error = False
#        first_cigarop = read.cigartuples[0][0]
#        if first_cigarop != 0:
#            if first_cigarop not in (1, 4):
#                error = True
#            else:
#                if read.cigartuples[1][0] != 0:
#                    error = True
#        if error:
#            raise_err(read)
#
#    def get_pileup_range(readlist):
#        start0s = list()
#        end0s = list()
#        for read in readlist:
#            start0s.append(read.reference_start)
#            end0s.append(read.reference_end)
#        start0 = min(start0s)
#        end0 = max(end0s)
#
#        return range(start0, end0)
#
#    def check_overlaps_pileup_range(read, pileup_range):
#        if read.cigartuples[-1][0] in readhandler.CIGAROPS_QUERYONLY:
#            overlaps_upstream = (read.reference_end >= pileup_range.start)
#        else:
#            overlaps_upstream = (read.reference_end > pileup_range.start)
#
#        overlaps_downstream = (read.reference_start < pileup_range.stop)
#
#        return overlaps_upstream and overlaps_downstream
#
#    def handle_leading_insclip(read):
#        leading_insclip_len = 0
#        for cigartup_idx, (cigarop, cigarlen) in enumerate(read.cigartuples):
#            if cigarop in (1, 4):
#                leading_insclip_len += cigarlen
#            else:
#                break
#
#        if leading_insclip_len == 0:
#            leading_insseq = None
#        else:
#            leading_insseq = read.query_sequence[:leading_insclip_len]
#
#        query_idx = leading_insclip_len
#
#        return leading_insseq, cigartup_idx, query_idx
#
#    def handle_last_match(
#        read, arr, cigarlen, query_idx, arr_col_idx, arr_row_idx, cigartup_idx
#    ):
#        sl_query = slice(query_idx, query_idx + cigarlen)
#        seq_buffer = tuple(read.query_sequence[sl_query])
#        sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen)
#        arr[arr_row_idx, sl_arr_col] = seq_buffer
#        cigartup_idx += 1
#
#        return cigartup_idx
#
#    def add_matches_before_last(
#        read, arr, cigarlen, query_idx, arr_col_idx, arr_row_idx
#    ):
#        sl_query = slice(query_idx, query_idx + cigarlen - 1)
#        seq = tuple(read.query_sequence[sl_query])
#        query_idx += cigarlen - 1
#
#        sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen - 1)
#        arr[arr_row_idx, sl_arr_col] = seq
#        arr_col_idx += cigarlen - 1
#
#        return query_idx, arr_col_idx
#
#    def handle_del_after_match(
#        read, arr, query_idx, arr_col_idx, arr_row_idx, next_cigarlen
#    ):
#        seq = read.query_sequence[query_idx]
#        query_idx += 1
#        arr[arr_row_idx, arr_col_idx] = seq
#        arr_col_idx += 1
#
#        sl_arr_col = slice(arr_col_idx, arr_col_idx + next_cigarlen)
#        arr[arr_row_idx, sl_arr_col] = del_char
#        arr_col_idx += next_cigarlen
#
#        return query_idx, arr_col_idx
#
#    def handle_insclip_after_match(
#        read, arr, query_idx, arr_col_idx, arr_row_idx, next_cigarlen
#    ):
#        match_seq = read.query_sequence[query_idx]
#        query_idx += 1
#        insseq = read.query_sequence[query_idx : (query_idx + next_cigarlen)]
#        query_idx += next_cigarlen
#        seq = match_seq + f"({insseq})"
#
#        arr[arr_row_idx, arr_col_idx] = seq
#        arr_col_idx += 1
#
#        return query_idx, arr_col_idx
#
#    def add_seqs(read, arr, cigartup_idx, query_idx, arr_row_idx, arr_col_idx):
#        while True:
#            if cigartup_idx == len(read.cigartuples):
#                break
#
#            cigarop, cigarlen = read.cigartuples[cigartup_idx]
#            if cigarop != 0:
#                raise_err(read)
#
#            if cigartup_idx == len(read.cigartuples) - 1:
#                cigartup_idx = handle_last_match(
#                    read,
#                    arr,
#                    cigarlen,
#                    query_idx,
#                    arr_col_idx,
#                    arr_row_idx,
#                    cigartup_idx,
#                )
#                continue
#            else:
#                if cigarlen > 1:
#                    query_idx, arr_col_idx = add_matches_before_last(
#                        read, arr, cigarlen, query_idx, arr_col_idx, arr_row_idx
#                    )
#
#                next_cigarop, next_cigarlen = read.cigartuples[cigartup_idx + 1]
#                if next_cigarop == 2:  # deletion
#                    query_idx, arr_col_idx = handle_del_after_match(
#                        read, arr, query_idx, arr_col_idx, arr_row_idx, next_cigarlen
#                    )
#                elif next_cigarop in (1, 4):
#                    query_idx, arr_col_idx = handle_insclip_after_match(
#                        read, arr, query_idx, arr_col_idx, arr_row_idx, next_cigarlen
#                    )
#
#                cigartup_idx += 2
#                continue
#
#    def add_leading_insseq(read, leading_insseq, arr, arr_row_idx, arr_col_idx_init):
#        if leading_insseq is not None:
#            old_seq = arr[arr_row_idx, arr_col_idx_init]
#            new_seq = f"(l+{leading_insseq})" + old_seq
#            arr[arr_row_idx, arr_col_idx_init] = f"({leading_insseq})" + old_seq
#
#    # main
#    chrom = readlist[0].reference_name
#    if pileup_range is None:
#        pileup_range = get_pileup_range(readlist)
#
#    ncol = len(pileup_range)
#    nrow = len(readlist)
#    #arr = np.empty((nrow, ncol), dtype=object)
#    arr = np.full((nrow, ncol), None, dtype=object)
#
#    for arr_row_idx, read in enumerate(
#        read for read in readlist
#        if check_overlaps_pileup_range(read, pileup_range)
#    ):
#        try:
#            arr_col_idx_init = read.reference_start - pileup_range.start
#            arr_col_idx = arr_col_idx_init
#            leading_insseq, cigartup_idx, query_idx = handle_leading_insclip(read)
#            add_seqs(read, arr, cigartup_idx, query_idx, arr_row_idx, arr_col_idx)
#            add_leading_insseq(read, leading_insseq, arr, arr_row_idx, arr_col_idx_init)
#        except Exception as exc:
#            try:
#                cigar_sanitycheck(read)
#            except Exception as exc_cigarpattern:
#                raise exc_cigarpattern from exc
#            else:
#                raise exc
#
#    if as_array:
#        if return_range:
#            return (arr, pileup_range)
#        else:
#            return arr
#    else:
#        uids = [readhandler.get_uid(read) for read in readlist]
#        if len(uids) != len(set(uids)):
#            reads_as_string = '\n'.join(x.to_string() for x in readlist)
#            raise Exception(f'Duplicate read uid. Input reads are as following:\n{reads_as_string}')
#
#        df = pd.DataFrame(arr, columns=list(pileup_range), index=uids)
#        pileup = Pileup(df=df, uids=uids, chrom=chrom)
#        if return_range:
#            return (pileup, pileup_range)
#        else:
#            return pileup
